from deposit_gui.dgui.dcdialogs import DCDialogs
from deposit_gui.dgui.dsave_as_postgres_frame import DSaveAsPostgresFrame
from deposit_gui.dialogs.dialog_connect import DialogConnect
from deposit_gui.dialogs.dialog_import_source import DialogImportSource
from deposit_gui.dialogs.dialog_import_store import DialogImportStore
from deposit_gui.dialogs.dialog_import_external import DialogImportExternal
from deposit_gui.dialogs.dialog_add_relation import DialogAddRelation
from deposit_gui.dialogs.dialog_about import DialogAbout
from deposit import Store

from PySide2 import (QtWidgets, QtCore, QtGui)
from natsort import (natsorted)
import os

class CDialogs(DCDialogs):
	
	def __init__(self, cmain, cview):
		
		DCDialogs.__init__(self, cmain, cview)
		
		self.view = cview._view
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot()
	def on_clear_recent(self):
		
		self.view.clear_recent_connections()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	
	
	# ---- Dialogs
	# ------------------------------------------------------------------------
	'''
	open(name, *args, **kwargs)
	
	Implement set_up_[name], process_[name] and cancel_[name] for each dialog:
	
	def set_up_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		dialog = QtWidgets.QDialog
		
		dialog.set_title(name)
		dialog.set_frame(frame = QtWidget)
		dialog.get_frame()
		dialog.set_button_box(ok: bool, cancel: bool): set if OK and Cancel buttons are visible
		
	def process_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		process dialog after OK has been clicked
	
	def cancel_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		handle dialog after cancel has been clicked
	'''
	
	def set_up_Connect(self, dialog, *args, **kwargs):
		
		connect_frame = DialogConnect(dialog)
		connect_frame.set_recent_dir(self.view.get_recent_dir())
		connect_frame.set_recent_connections(self.view.get_recent_connections())
		connect_frame.signal_clear_recent.connect(self.on_clear_recent)
	
	def process_Connect(self, dialog, *args, **kwargs):
		
		self.cmain.cmodel.load(**dialog.get_data())
	
	def cancel_Connect(self, dialog, *args, **kwargs):
		
		datasource = self.cmain.cmodel.get_datasource()
		if not datasource.is_valid():
			self.cmain.cmodel.load(datasource = "Memory")
	
	
	def set_up_SaveAsPostgres(self, dialog, *args, **kwargs):
		
		frame = DSaveAsPostgresFrame(dialog)
		frame.set_recent_connections(self.view.get_recent_connections())
	
	def process_SaveAsPostgres(self, dialog, *args, **kwargs):
		
		data = dialog.get_data()
		datasource = data["datasource"]
		self.cmain.cmodel.set_local_folder(datasource.get_local_folder())
		self.cmain.cmodel.save(**data)
		
		reply = QtWidgets.QMessageBox.question(self.cmain.cview._view, 
			"Load Database?", 
			"Load database from <b>%s</b>" % (datasource.get_name()),
		)
		if reply == QtWidgets.QMessageBox.Yes:
			self.cmain.cmodel.load(**data)
	
	
	def set_up_ImportStoreSource(self, dialog):
		
		connect_frame = DialogImportSource(dialog)
		connect_frame.set_recent_dir(self.view.get_recent_dir())
		connect_frame.set_recent_connections(self.view.get_recent_connections())
	
	def process_ImportStoreSource(self, dialog):
		
		self.cmain.cview.progress.show("Loading")
		store = Store(keep_temp = True)
		store.load(progress = self.cmain.cview.progress, **dialog.get_data())
		self.cmain.cview.progress.stop()
		if not store.get_objects():
			return
		self.open("ImportStore", store)
	
	def set_up_ImportStore(self, dialog, store):
		
		classes = store.get_class_names(ordered = True)
		descriptors = store.get_descriptor_names()
		classes = [name for name in classes if name not in descriptors]
		frame = DialogImportStore(dialog, store.get_objects(), classes)
		dialog.set_frame(frame)
	
	def process_ImportStore(self, dialog, store):
		
		unique = dialog.get_frame().get_unique()
		self.cmain.cmodel.import_store(store, unique)
	
	
	def set_up_ImportExternal(self, dialog, external_frame):
		
		targets = external_frame.get_targets()  # {column_idx: (class_name, descriptor_name), ...}
		classes = set()
		for class_name, _ in targets.values():
			cls = self.cmain.cmodel.get_class(class_name)
			if cls:
				classes.add(cls)
		relations = set()
		for cls in classes:
			for cls2, label in cls.get_relations():
				if label.startswith("~"):
					continue
				if cls2 not in classes:
					continue
				relations.add((cls.name, label, cls2.name))
		classes = natsorted([cls.name for cls in classes])
		relations = natsorted(relations)
		
		n_rows = external_frame.get_row_count()
		frame = DialogImportExternal(dialog, n_rows, classes, relations)
		dialog.set_frame(frame)
	
	def process_ImportExternal(self, dialog, external_frame):
		
		frame = dialog.get_frame()
		unique = frame.get_unique()
		relations = frame.get_relations()
		n_rows = external_frame.get_row_count()
		targets = external_frame.get_targets()
		n_imported = self.cmain.cmodel.import_data(external_frame.get_data, n_rows, targets, relations, unique)
		self.cmain.cview.show_information("Import Successful", "Imported %d Objects" % (n_imported))
	
	
	def set_up_ConfirmLoad(self, dialog, url):
		
		dialog.set_title("Load Database?")
		dialog.setModal(True)
		dialog.set_button_box(True, True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		frame.layout().addWidget(QtWidgets.QLabel("Load database from <b>%s</b>?" % (os.path.split(url)[-1])))
		
		dialog.set_frame(frame)
	
	def process_ConfirmLoad(self, dialog, url):
		
		self.cmain.cmodel.load(url = url)
	
	
	def set_up_AddClass(self, dialog, parent_cls):
		
		dialog.set_title("Add Class")
		dialog.setModal(True)
		dialog.set_button_box(True, True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		frame.name_edit = QtWidgets.QLineEdit()
		frame.name_edit.setPlaceholderText("Class name")
		caption = ""
		if parent_cls is not None:
			caption = "As subclass of %s" % (parent_cls.name)
		frame.as_subclass = QtWidgets.QCheckBox(caption)
		frame.layout().addWidget(frame.name_edit)
		if parent_cls is not None:
			frame.layout().addWidget(frame.as_subclass)
			
		dialog.set_frame(frame)
	
	def process_AddClass(self, dialog, parent_cls):
		
		frame = dialog.get_frame()
		name = frame.name_edit.text()
		as_subclass = frame.as_subclass.isChecked()
		if not name:
			return
		cls = self.cmain.cmodel.add_class(name)
		if as_subclass:
			parent_cls.add_subclass(cls)
	
	
	def set_up_AddDescriptor(self, dialog, cls):
		
		dialog.set_title("Add Descriptor")
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		frame.name_edit = QtWidgets.QLineEdit()
		frame.name_edit.setPlaceholderText("Descriptor name")
		frame.layout().addWidget(frame.name_edit)
		
		dialog.set_frame(frame)
	
	def process_AddDescriptor(self, dialog, cls):
		
		frame = dialog.get_frame()
		name = frame.name_edit.text()
		if not name:
			return
		self.cmain.cmodel.get_class(cls).set_descriptor(name)
	
	
	def set_up_DelDescriptor(self, dialog, descr, cls):
		
		dialog.set_title("Remove Descriptor")
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		frame.layout().addWidget(QtWidgets.QLabel(
			"Remove descriptor <b>%s</b> from class <b>%s</b>?" % (descr.name, cls.name)
		))
		frame.class_only = QtWidgets.QCheckBox("Only remove class descriptor (not from objects)")
		frame.layout().addWidget(frame.class_only)
		
		dialog.set_frame(frame)
	
	def process_DelDescriptor(self, dialog, descr, cls):
		
		frame = dialog.get_frame()
		class_only = frame.class_only.isChecked()
		self.cmain.cmodel.del_class_descriptor(descr, cls, class_only)
	
	
	def set_up_RenameClass(self, dialog, cls, is_descriptor, parent_cls):
		
		dialog.set_title("Rename Descriptor" if is_descriptor else "Rename Class")
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		frame.name_edit = QtWidgets.QLineEdit()
		frame.name_edit.setText(cls.name)
		caption = ""
		if is_descriptor and (parent_cls is not None):
			caption = "Only rename Descriptor of class %s" % (parent_cls.name)
		frame.parent_only = QtWidgets.QCheckBox(caption)
		frame.layout().addWidget(frame.name_edit)
		if is_descriptor and (parent_cls is not None):
			frame.layout().addWidget(frame.parent_only)
			
		dialog.set_frame(frame)
	
	def process_RenameClass(self, dialog, cls, is_descriptor, parent_cls):
		
		frame = dialog.get_frame()
		name = frame.name_edit.text()
		parent_only = frame.parent_only.isChecked()
		if not name:
			return
		if parent_only:
			self.cmain.cmodel.rename_class_descriptor(cls, parent_cls, name)
		else:
			self.cmain.cmodel.rename_class(cls, name)
	
	
	def set_up_DelClass(self, dialog, classes):
		
		dialog.set_title("Remove Class")
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		
		names = [cls.name for cls in classes]
		frame.layout().addWidget(QtWidgets.QLabel(
			"Remove the following class%s?" % ("es" if (len(names) > 1) else "")
		))
		frame.layout().addWidget(QtWidgets.QLabel("<b>%s</b>" % ", ".join(names)))
		caption = ""
		frame.superclasses = set()
		for cls in classes:
			frame.superclasses.update(set(cls.get_superclasses()))
		frame.del_superclasses = QtWidgets.QCheckBox("Only remove from superclasses?")
		if frame.superclasses:
			frame.layout().addWidget(frame.del_superclasses)
			
		dialog.set_frame(frame)
	
	def process_DelClass(self, dialog, classes):
		
		frame = dialog.get_frame()
		del_superclasses = frame.del_superclasses.isChecked()
		if del_superclasses:
			for supcls in frame.superclasses:
				for cls in classes:
					supcls.del_subclass(cls)
		else:
			for cls in classes:
				self.cmain.cmodel.del_class(cls)
	
	
	def set_up_AddSavedQuery(self, dialog):
		
		querystr = self.cmain.cquerytoolbar.get_query_text()
		
		dialog.set_title("Add Query")
		dialog.setModal(True)
		dialog.set_button_box(True, True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		frame.title_edit = QtWidgets.QLineEdit()
		frame.title_edit.setPlaceholderText("Title")
		frame.querystr_edit = QtWidgets.QPlainTextEdit()
		if querystr:
			frame.querystr_edit.setPlainText(querystr)
		else:
			frame.querystr_edit.setPlaceholderText("Query String")
		frame.layout().addWidget(frame.title_edit)
		frame.layout().addWidget(frame.querystr_edit)
		
		dialog.set_frame(frame)
	
	def process_AddSavedQuery(self, dialog):
		
		frame = dialog.get_frame()
		title = frame.title_edit.text()
		querystr = frame.querystr_edit.toPlainText()
		if title and querystr:
			self.cmain.cmodel.add_saved_query(title, querystr)
	
	
	def set_up_EditSavedQuery(self, dialog, title):
		
		querystr = self.cmain.cmodel.get_saved_query(title)
		
		dialog.set_title("Edit Query")
		dialog.setModal(True)
		dialog.set_button_box(True, True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		frame.title_edit = QtWidgets.QLineEdit()
		frame.title_edit.setText(title)
		frame.querystr_edit = QtWidgets.QPlainTextEdit()
		frame.querystr_edit.setPlainText(querystr)
		frame.layout().addWidget(frame.title_edit)
		frame.layout().addWidget(frame.querystr_edit)
		
		dialog.set_frame(frame)
		
	
	def process_EditSavedQuery(self, dialog, title):
		
		frame = dialog.get_frame()
		title_new = frame.title_edit.text()
		querystr = frame.querystr_edit.toPlainText()
		if title_new and querystr:
			if title_new != title:
				self.cmain.cmodel.del_saved_query(title)
			self.cmain.cmodel.add_saved_query(title_new, querystr)
	
	def set_up_RemoveSavedQuery(self, dialog, titles):
		
		dialog.set_title("Remove %s" % ("Queries" if len(titles) > 1 else "Query"))
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		
		frame.layout().addWidget(QtWidgets.QLabel(
			"Remove the following %s?" % ("queries" if len(titles) > 1 else "query")
		))
		frame.layout().addWidget(QtWidgets.QLabel("".join(["<p><b>%s</b></p>" % (title) for title in titles])))
		
		dialog.set_frame(frame)
	
	def process_RemoveSavedQuery(self, dialog, titles):
		
		for title in titles:
			self.cmain.cmodel.del_saved_query(title)
	
	
	def set_up_DelObjects(self, dialog, objects):
		
		dialog.set_title("Remove Object%s" % ("s" if len(objects) > 1 else ""))
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		
		frame.layout().addWidget(QtWidgets.QLabel(
			"Remove the following object%s?" % ("s" if (len(objects) > 1) else "")
		))
		frame.layout().addWidget(QtWidgets.QLabel("<b>%s</b>" % ", ".join(sorted([str(obj.id) for obj in objects]))))
		caption = ""
		
		dialog.set_frame(frame)
	
	def process_DelObjects(self, dialog, objects):
		
		self.cmain.cmodel.del_objects(objects)
	
	
	def set_up_AddRelation(self, dialog, elements, label):
		
		is_class = (elements[0].__class__.__name__ == "DClass")
		dialog.set_frame(DialogAddRelation(dialog, self.cmain, elements, label, is_class))
	
	def process_AddRelation(self, dialog, elements, label):
		
		frame = dialog.get_frame()
		label = frame.get_label()
		targets = frame.get_targets()
		if (not label) or (not targets):
			return
		for src in elements:
			for tgt in targets:
				src.add_relation(tgt, label)
	
	
	def set_up_DelRelation(self, dialog, relations):
		
		dialog.set_title("Remove Relation%s" % ("s" if len(relations) > 1 else ""))
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(300)
		frame.setLayout(QtWidgets.QVBoxLayout())
		
		frame.layout().addWidget(QtWidgets.QLabel(
			"Remove the selected relation%s?" % ("s" if (len(relations) > 1) else "")
		))
		
		dialog.set_frame(frame)
	
	def process_DelRelation(self, dialog, relations):
		
		for src, tgt, label in relations:
			src.del_relation(tgt, label)
	
	
	def set_up_FieldCalculator(self, dialog, rows, descriptor_name):
		
		dialog.set_title("About Deposit")
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(600)
		frame.setLayout(QtWidgets.QHBoxLayout())
		frame.target_edit = QtWidgets.QLineEdit()
		if descriptor_name is None:
			frame.target_edit.setPlaceholderText("Descriptor name")
		else:
			frame.target_edit.setText(descriptor_name)
		frame.target_edit.setFixedWidth(100)
		frame.expr_edit = QtWidgets.QPlainTextEdit()
		frame.expr_edit.setPlaceholderText(
		'''Python expression e.g.:

Class1.Descr1 + [Class 1].[Descr 2]
Class1.Descr1.filename if isinstance(Class1.Descr1, DResource) else None
DGeometry('POINT(10 20)')
DGeometry('POINT', [10, 20], 4326)  # 4326 = SRID
DResource('C:/folder/' + Class1.Descr1.filename)
DDateTime('2022-07-20T19:31:00.729558')
		''')
		label = QtWidgets.QLabel("=")
		frame.layout().addWidget(frame.target_edit)
		frame.layout().addWidget(label)
		frame.layout().addWidget(frame.expr_edit)
		
		frame.layout().setAlignment(frame.target_edit, QtCore.Qt.AlignTop)
		frame.layout().setAlignment(label, QtCore.Qt.AlignTop)
		
		dialog.set_frame(frame)
	
	def process_FieldCalculator(self, dialog, rows, descriptor_name):
		
		frame = dialog.get_frame()
		target = frame.target_edit.text().strip()
		expr = frame.expr_edit.toPlainText().strip()
		if target and expr:
			self.cmain.cmodel.field_calculator(target, expr, rows)
	
	
	def set_up_About(self, dialog):
		
		dialog.set_title("About Deposit")
		dialog.set_button_box(True, False)
		dialog.setModal(True)
		dialog.set_frame(DialogAbout())




from PySide2 import (QtWidgets, QtCore, QtGui)

class DConnectTabRecent(QtWidgets.QFrame):
	
	def __init__(self, parent):
		
		QtWidgets.QFrame.__init__(self, parent)
		
		self.parent = parent
		
		self.setLayout(QtWidgets.QHBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self.left = QtWidgets.QFrame()
		self.left.setLayout(QtWidgets.QVBoxLayout())
		self.left.layout().setContentsMargins(0, 0, 0, 0)
		self.right = QtWidgets.QFrame()
		self.right.setLayout(QtWidgets.QVBoxLayout())
		self.right.layout().setContentsMargins(0, 0, 0, 0)
		
		self.recent_list = QtWidgets.QListWidget()
		self.recent_list.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
		self.recent_list.itemSelectionChanged.connect(self.on_selected)
		self.recent_list.activated.connect(self.on_connect)
		
		self.connect_button = QtWidgets.QPushButton(self.parent.connect_caption())
		self.connect_button.clicked.connect(self.on_connect)
		
		self.clear_recent_button = QtWidgets.QPushButton("Clear Recent")
		self.clear_recent_button.clicked.connect(self.on_clear_recent)
		
		self.left.layout().addWidget(self.recent_list)
		self.left.layout().addWidget(self.connect_button)
		self.layout().addWidget(self.left)
		
		logo = self.parent.logo()
		if logo:
			self.right.layout().addWidget(logo)
			self.right.layout().addWidget(self.clear_recent_button)
			self.layout().addWidget(self.right)
		
		self.update()
	
	def set_recent_connections(self, data):
		# data = [[url], [identifier, connstr], ...]
		
		for row in data:
			name = None
			if len(row) == 1:
				url = row[0]
				name = "File: %s" % (url)
			elif len(row) == 2:
				identifier, connstr = row
				name = "DB: %s (%s)" % (identifier, connstr.split("@")[-1].split("?")[0])
			if not name:
				continue
			item = QtWidgets.QListWidgetItem(name)
			item.setData(QtCore.Qt.UserRole, row)
			self.recent_list.addItem(item)
		
		self.update()
	
	def has_items(self):
		
		return self.recent_list.count() > 0
	
	def update(self):
		
		self.connect_button.setEnabled(len(self.recent_list.selectedItems()) > 0)
	
	def on_selected(self):
		
		self.update()
	
	@QtCore.Slot()
	def on_connect(self):
		
		item = self.recent_list.currentItem()
		if not item:
			return
		row = item.data(QtCore.Qt.UserRole)
		if len(row) == 1:
			self.parent.on_connect(url = row[0])
			return
		
		if len(row) == 2:
			identifier, connstr = row
			self.parent.on_connect(identifier = identifier, connstr = connstr)
			return
	
	@QtCore.Slot()
	def on_clear_recent(self):
		
		self.recent_list.clear()
		self.parent.signal_clear_recent.emit()




'''
VActions(AbstractSubview)
	.toolbar
		.signal_triggered = Signal(name)
		.signal_combo_changed = Signal(name, text)
		.populate(tools)
		.get_names()
		.update_tool(name, data, child_window = False, clear_combo_text = False)
		.set_toolbar_visible(name, state)
		.get_state(name)
		.get_combo_value(name)
		.set_combo_value(name, text)
	.menubar
		same as toolbar

'''

from deposit_gui.dgui.abstract_subview import AbstractSubview
from deposit_gui.dgui.dtool_bar import DToolBar
from deposit_gui.dgui.dmenu_bar import DMenuBar

from PySide2 import (QtCore)

class DVActions(AbstractSubview):
	
	def __init__(self, vmain) -> None:
		
		AbstractSubview.__init__(self, vmain)
		
		# implement each action behaviour via lap.controller.cactions
		
		self.toolbar = DToolBar(self.vmain)
		self.menubar = DMenuBar(self.vmain)



'''
Generic class to be re-used also by Deposit in the future
'''

from PySide2 import (QtWidgets)

class DStatusBar(QtWidgets.QStatusBar):
	
	def __init__(self, view):
		
		QtWidgets.QStatusBar.__init__(self, view)
		
	def message(self, text):
		
		self.showMessage(text)




from deposit_gui.dgui.abstract_subview import AbstractSubview

class AbstractMDIAreaFrame(AbstractSubview):
	
	def __init__(self):
		
		AbstractSubview.__init__(self)
	
	def title(self):
		# re-implement
		
		return "MDIAreaFrame"
	
	def icon(self):
		# re-implement
		
		return "dep_cube.svg"
	
	def on_close(self):
		# re-implement
		
		pass
	
	def on_deactivate(self):
		# re-implement
		
		pass



from deposit_gui.view.vusertools_elements.user_elements.user_tool import (UserTool)

from PySide2 import (QtCore)

class Query(UserTool):
	
	def __init__(self, label, value, vusertools):
		
		self.value = value
		
		UserTool.__init__(self, label, vusertools)
		
		self.setIcon(self._vusertools.get_icon("query.svg"))
		self.setToolTip("Query: %s" % self.label)
		
		self._vusertools._cmodel.cmain.signal_selection_changed.connect(self.update)
		
		self.update()
	
	@QtCore.Slot()
	def update(self):
		
		if self._vusertools is None:
			return
		if (self._vusertools.SELECTED_STR not in self.value) or (self._vusertools.get_selected_id() is not None):
			self.setEnabled(True)
		else:
			self.setEnabled(False)
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self._vusertools.open_query(self)
	
	def to_markup(self):
		
		return UserTool.to_markup(self) + '''
<QueryString>%s</>
		''' % (self.value)
	
	def to_dict(self):
		
		return dict(
			value = self.value,
			**UserTool.to_dict(self),
		)




from deposit_gui.view.vmdiarea_frames.abstract_mdiarea_frame import AbstractMDIAreaFrame
from deposit_gui.dgui.dgraph_view import (DGraphView, NodeWithAttributes, NodeWithSimpleAttributes)

from PySide2 import (QtWidgets, QtCore, QtGui)
from pathlib import Path
import os

class ClassGraphFrame(AbstractMDIAreaFrame, QtWidgets.QFrame):
	
	signal_class_selected = QtCore.Signal(list)		# [Class, ...]
	signal_relation_selected = QtCore.Signal(list)		# [(Source, Target, label), ...]
	signal_activated = QtCore.Signal(object)	# Class
	
	def __init__(self, cmodel, cview):
		
		AbstractMDIAreaFrame.__init__(self)
		QtWidgets.QFrame.__init__(self)
		
		self._graph = ClassGraph(self, cmodel, cview)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().setSpacing(0)
		self.layout().addWidget(self._graph)
	
	def title(self):
		
		return "Class Relations"
	
	def icon(self):
		# re-implement
		
		return "classes_graph.svg"
	
	def update_classes(self):
		
		self._graph.update_classes()
	
	def on_deactivate(self):
		
		self._graph._graph_view.deselect_all()

class ClassGraph(QtWidgets.QMainWindow):
	
	def __init__(self, frame, cmodel, cview):
		
		QtWidgets.QMainWindow.__init__(self)
		
		self._frame = frame
		self._cmodel = cmodel
		self._cview = cview
		
		self._actions = {} # {name: QAction, ...}
		self._positions = {}  # {node_id: (x, y), ...}
		self._show_descriptors = False
		
		self._graph_view = DGraphView()
		self._graph_view.signal_node_activated.connect(self.on_activated)
		self._graph_view.signal_selected.connect(self.on_selected)
		
		central_widget = QtWidgets.QWidget(self)
		central_widget.setLayout(QtWidgets.QVBoxLayout())
		central_widget.layout().setContentsMargins(0, 0, 0, 0)
		self.setCentralWidget(central_widget)
		central_widget.layout().addWidget(self._graph_view)
		
		self.toolbar = self.addToolBar("Graph")
		actions = [
			["descriptor_view", "Show Descriptors", "attributes.svg"],
			["reset_layout", "Re-arrange Layout", "geometry.svg"],
			["#separator", None, None],
			["zoom_in", "Zoom In", "zoom_in.svg"],
			["zoom_out", "Zoom Out", "zoom_out.svg"],
			["zoom_reset", "Zoom Reset", "zoom_reset.svg"],
			["#separator", None, None],
			["save_pdf", "Save As PDF", "save_pdf.svg"],
		]
		
		for name, text, icon in actions:
			if name == "#separator":
				self.toolbar.addSeparator()
			else:
				self._actions[name] = QtWidgets.QAction(self._frame.get_icon(icon), text, self)
				self._actions[name].setData(name)
				self.toolbar.addAction(self._actions[name])
		
		self._actions["descriptor_view"].setCheckable(True)
		
		self.toolbar.actionTriggered.connect(self.on_tool_triggered)
		
		self.populate_graph()
	
	def update_classes(self):
		
		self.populate_graph()
	
	def populate_graph(self):
		
		nodes = []
		edges = []
		descriptors = self._cmodel.get_descriptor_names()
		for cls in self._cmodel.get_classes():
			if cls.name in descriptors:
				continue
			if self._show_descriptors:
				descriptors = []
				for name in cls.get_descriptor_names(ordered = True):
					descriptors.append((name, ""))
				nodes.append(NodeWithAttributes(cls.name, cls.name, descriptors))
			else:
				nodes.append(NodeWithSimpleAttributes(cls.name, cls.name, [("name", cls.name)]))
			for cls2, label in cls.get_object_relations(direct_only = True):
				if label.startswith("~"):
					continue
				if cls2 == cls:
					continue
				edges.append([cls.name, cls2.name, label, "darkGray"])
			for cls2, label in cls.get_relations():
				if label.startswith("~"):
					continue
				if cls2 == cls:
					continue
				edge = [cls.name, cls2.name, label, "darkGray"]
				if edge not in edges:
					edges.append(edge)
		self._graph_view.populate(nodes = nodes, edges = edges, positions = self._positions)
	
	@QtCore.Slot(object)
	def on_tool_triggered(self, action):
		
		fnc_name = "on_%s" % str(action.data())
		if hasattr(self, fnc_name):
			getattr(self, fnc_name)()
	
	@QtCore.Slot(object)
	def on_activated(self, name):
		
		self._frame.signal_activated.emit(self._cmodel.get_class(name))
	
	@QtCore.Slot()
	def on_selected(self):
		
		nodes, edges = self._graph_view.get_selected()
		self._frame.signal_class_selected.emit([self._cmodel.get_class(name) for name in nodes])
		if nodes:
			return
		relations = []
		for source, target, label in edges:
			relations.append((self._cmodel.get_class(source), self._cmodel.get_class(target), label))
		self._frame.signal_relation_selected.emit(relations)
	
	
	# Toolbar actions
	
	def on_descriptor_view(self):
		
		self._show_descriptors = self._actions["descriptor_view"].isChecked()
		self.populate_graph()
	
	def on_reset_layout(self):
		
		self._positions = {}
		self.populate_graph()
	
	def on_zoom_in(self):
		
		self._graph_view.scale(1.1, 1.1)
	
	def on_zoom_out(self):
		
		self._graph_view.scale(0.9, 0.9)
	
	def on_zoom_reset(self):
		
		self._graph_view.reset_zoom()
	
	def on_save_pdf(self):
		
		filename = "%s_schema.pdf" % (self._cmodel.get_datasource_name())
		recent_dir = self._cview.get_recent_dir()
		if not recent_dir:
			recent_dir = str(Path.home())
		path = os.path.join(recent_dir, filename)
		path, _ = QtWidgets.QFileDialog.getSaveFileName(None, "Save As Adobe PDF", path, "Adobe PDF (*.pdf)")
		if not path:
			return
		self._graph_view.save_pdf(path)
		self._cview.set_recent_dir(os.path.dirname(path))





from deposit_gui.view.vusertools_elements.dialog.dialog_form import (DialogForm)

from deposit.utils.fnc_serialize import (value_to_str)

from PySide2 import (QtWidgets, QtCore)
from collections import defaultdict
from natsort import natsorted

class DialogEntryForm(DialogForm):
	
	signal_submit = QtCore.Signal(dict, dict, list, set, dict)
	#	values, objects_existing, relations, unique, objects_loaded
	signal_unlink = QtCore.Signal(dict, list, set, list)
	#	objects_existing, relations, unique, obj_ids
	signal_remove = QtCore.Signal(dict, set)
	#	objects_existing, unique
	#		values = {cls: {idx: {descr: value, ...}, ...}, ...}
	#		objects_existing = {cls: {idx: obj_id, ...}, ...}
	#		relations = [[cls1, rel, cls2], ...]
	#		unique = set(cls, ...)
	
	def __init__(self, vusertools, form_tool, selected_id):
		
		DialogForm.__init__(self, vusertools, form_tool)
		
		self.selected_id = selected_id
		self.objects_loaded = {}
		
		if self._vusertools.entry_form_geometry is not None:
			self.setGeometry(self._vusertools.entry_form_geometry)
		else:
			self.setMinimumWidth(600)
		
		self.overrideWindowFlags(QtCore.Qt.Window)
		
		button_remove = QtWidgets.QPushButton("Delete")
		button_remove.clicked.connect(self.on_remove)
		self.button_frame.layout().addWidget(button_remove)
		
		if self.selected_id is not None:
			self.populate()
	
	def find_relations(self):
		# find all possible relations
		# returns frames, framesets, relations
		# frames = [DialogFrame(), ...]
		# framesets = [[DialogFrame(), ...], ...]
		# relations = [[cls1, rel, cls2], ...]
		
		relations = []  # [[cls1, rel, cls2], ...]
		classes = set()
		frames, framesets = self.frames()
		for frame in frames:
			classes.add(frame.dclass)
		for frameset in framesets:
			for frame in frameset:
				classes.add(frame.dclass)
		for cls1 in classes:
			cls1 = self._vusertools._cmodel.get_class(cls1)
			if cls1 is None:
				continue
			for cls2, rel in cls1.get_relations():
				if rel.startswith("~"):
					continue
				if cls2 == cls1:
					continue
				if cls2.name not in classes:
					continue
				relations.append([cls1.name, rel, cls2.name])
		relations = natsorted(relations)
		
		return frames, framesets, relations
	
	def populate(self):
		
		self.clear()
		
		# find all possible relations
		frames, framesets, relations = self.find_relations()
		collect = []
		for cls1, rel, cls2 in relations:
			rel_rev = self._vusertools._cmodel.reverse_relation(rel)
			collect.append([cls2, rel_rev, cls1])
		relations += collect
		
		# find classes from multi groups
		multi_classes = defaultdict(set)  # {cls: set(group, ...), ...}
		for group in self.multigroups():
			for frameset in group.framesets():
				for frame in frameset:
					multi_classes[frame.dclass].add(group)
		
		# get all relevant objects
		objects = defaultdict(list)  # {class: [object, ...], ...}
		obj = self._vusertools._cmodel.get_object(self.selected_id)
		if obj is None:
			return
		for cls in obj.get_class_names():
			objects[cls].append(obj)
		while True:
			found = False
			for cls1, rel, cls2 in relations:
				if cls1 in objects:
					for obj1 in objects[cls1]:
						for obj2, rel in obj1.get_relations():
							classes2 = obj2.get_class_names()
							if set(obj1.get_class_names()).intersection(classes2):
								continue
							if (cls2 in classes2) and ((cls2 in multi_classes) or (cls2 not in objects)) and (obj2 not in objects[cls2]):
								objects[cls2].append(obj2)
								found = True
			if not found:
				break
		
		# collect data from objects
		data = {}  # {cls: {obj_id: {descr: value, ...}, ...}, ...}
		for cls in objects:
			data[cls] = {}
			for obj in objects[cls]:
				data[cls][obj.id] = {}			
				for descr in obj.get_descriptors():
					value = value_to_str(obj.get_descriptor(descr))
					if value is None:
						continue
					data[cls][obj.id][descr.name] = value
		
		# fill single entry frames
		for frame in frames:
			if frame.dclass not in data:
				continue
			obj_id = next(iter(data[frame.dclass]))
			if frame.descriptor in data[frame.dclass][obj_id]:
				frame.set_value(data[frame.dclass][obj_id][frame.descriptor], obj_id)
		
		# fill multi entry fields
		group_ids = defaultdict(set)  # {obj_id: set(group, ...), ...}
		frameset_values = defaultdict(lambda: defaultdict(dict))  # {obj_id: {cls: {descriptor: value, ...}, ...}, ...}
		for cls in multi_classes:
			if cls not in data:
				continue
			for obj_id in data[cls]:
				group_ids[obj_id].update(multi_classes[cls])
				for descr in data[cls][obj_id]:
					value = data[cls][obj_id][descr]
					if value is not None:
						frameset_values[obj_id][cls][descr] = value
		for obj_id in frameset_values:
			for group in group_ids[obj_id]:
				frameset = group.framesets()[-1]
				for frame in frameset:
					if (frame.dclass in frameset_values[obj_id]) and (frame.descriptor in frameset_values[obj_id][frame.dclass]):
						frame.set_value(frameset_values[obj_id][frame.dclass][frame.descriptor], obj_id)
				group.add_entry()
		self.adjust_labels()
		
		_, objects_existing, _ = self.get_data()
		self.objects_loaded = dict([(cls, list(objects_existing[cls].values())) for cls in objects_existing])
	
	def get_data(self):
		
		# find frames, framesets and all possible relations
		frames, framesets, relations = self.find_relations()
		# frames = [DialogFrame(), ...]
		# framesets = [[DialogFrame(), ...], ...]
		# relations = [[cls1, rel, cls2], ...]
		
		# collect values from form
		values = defaultdict(lambda: defaultdict(dict))  # {cls: {idx: {descr: value, ...}, ...}, ...}
		objects_existing = defaultdict(dict)  # {cls: {idx: obj_id, ...}, ...}
		for frame in frames:
			value = frame.get_value()
			if value == "":
				value = None
			values[frame.dclass][0][frame.descriptor] = value
			obj_id = frame.get_object()
			if obj_id is not None:
				objects_existing[frame.dclass][0] = obj_id
		idx = -1
		for frameset in framesets:
			idx += 1
			for frame in frameset:
				value = frame.get_value()
				if value == "":
					value = None
				values[frame.dclass][idx][frame.descriptor] = value
				obj_id = frame.get_object()
				if obj_id is not None:
					objects_existing[frame.dclass][idx] = obj_id
		
		return values, objects_existing, relations
	
	def clear(self):
		
		for group in self.multigroups():
			group.clear()
		frames, _ = self.frames()
		for frame in frames:
			frame.set_value("")
	
	@QtCore.Slot(list)
	def on_entry_removed(self, obj_ids):
		
		if not obj_ids:
			return
		_, objects_existing, relations = self.get_data()
		
		self.signal_unlink.emit(objects_existing, relations, self.unique, obj_ids)
		self.clear()
		self.populate()
	
	@QtCore.Slot()
	def on_submit(self):
		
		values, objects_existing, relations = self.get_data()
		self.signal_submit.emit(values, objects_existing, relations, self.unique, self.objects_loaded)
		self.clear()
	
	@QtCore.Slot()
	def on_reset(self):
		
		self.clear()
	
	@QtCore.Slot()
	def on_remove(self):
		
		if not self.unique:
			return
		_, objects_existing, _ = self.get_data()  # {cls: {idx: obj_id, ...}, ...}
		self.selected_id = None
		self.clear()
		self.signal_remove.emit(objects_existing, self.unique)
		self.populate()
	
	def closeEvent(self, event):
		
		self._vusertools.entry_form_geometry = self.geometry()
		DialogForm.closeEvent(self, event)




from deposit_gui.dgui.abstract_subview import AbstractSubview

from PySide2 import (QtWidgets, QtCore, QtGui)

class AbstractOutsideFrame(AbstractSubview, QtWidgets.QMainWindow):
	
	def __init__(self):
		
		AbstractSubview.__init__(self)
		QtWidgets.QMainWindow.__init__(self)
		
	
	def title(self):
		# re-implement
		
		return "MDIAreaFrame"
	
	def icon(self):
		# re-implement
		
		return "dep_cube.svg"




from deposit_gui.view.vmdiarea_frames.abstract_mdiarea_frame import AbstractMDIAreaFrame
from deposit import externalsource as Externalsource
from deposit import DGeometry
from deposit.query.parse import (remove_bracketed_all, replace_bracketed)

from PySide2 import (QtWidgets, QtCore, QtGui)

class ExternalFrame(AbstractMDIAreaFrame, QtWidgets.QWidget):
	
	def __init__(self, source, *args, **kwargs):
		
		AbstractMDIAreaFrame.__init__(self)
		QtWidgets.QWidget.__init__(self)
		
		self.tabs = None
		self.sheet = None
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().setSpacing(0)
		self.setLayout(QtWidgets.QVBoxLayout())
		
		if not hasattr(Externalsource, source):
			raise Exception("External source not supported: %s" % (source))
		self._externalsource = getattr(Externalsource, source)()
		self._externalsource.load(*args, **kwargs)
		
		sheets = self._externalsource.sheets()
		if len(sheets) > 1:
			self.tabs = QtWidgets.QTabWidget()
			self.layout().addWidget(self.tabs)
			for sheet_name in sheets:
				self.tabs.addTab(
					ExternalSheet(
						ExternalHeader(self._externalsource, sheet_name),
						ExternalBody(self, self._externalsource, sheet_name)
					),
					sheet_name
				)
		else:
			self.sheet = ExternalSheet(
						ExternalHeader(self._externalsource, sheets[0]),
						ExternalBody(self, self._externalsource, sheets[0])
					)
			self.layout().addWidget(self.sheet)
	
	def title(self):
		
		if self._externalsource is None:
			return "External Source"
		return self._externalsource.get_name()
	
	def icon(self):
		
		return "dep_cube.svg"
	
	def get_current(self):
		
		if self.tabs is not None:
			return self.tabs.currentWidget()
		return self.sheet
	
	def get_current_header(self):
		
		return self.get_current().header
	
	def get_current_body(self):
		
		return self.get_current().body
	
	def get_targets(self):
		# returns {column_idx: (class_name, descriptor_name), ...}
		
		header = self.get_current_header()
		targets = {}
		for idx in range(self._externalsource.column_count(header.sheet_name())):
			substr, bracketed = remove_bracketed_all(header.item(0, idx).data(QtCore.Qt.DisplayRole))
			data = substr.split(".")
			if len(data) != 2:
				continue
			data = [replace_bracketed(name, bracketed).strip("[]") for name in data]
			targets[idx] = tuple(data)
			
		return targets
	
	def get_column_count(self):
		
		if self._externalsource is None:
			return 0
		return self._externalsource.column_count(self.get_current_header().sheet_name())
	
	def get_row_count(self):
		
		if self._externalsource is None:
			return 0
		return self._externalsource.row_count(self.get_current_header().sheet_name())
	
	def get_data(self, row, col):
		
		if self._externalsource is None:
			return None
		
		return self._externalsource.data(self.get_current_header().sheet_name(), row, col)
	
	def on_close(self):
		# re-implement
		
		pass
	
	def on_deactivate(self):
		# re-implement
		
		pass

class ExternalSheet(QtWidgets.QWidget):
	
	def __init__(self, header, body):
		
		self.header = header
		self.body = body
		
		QtWidgets.QWidget.__init__(self)
		
		self.verticalLayout = QtWidgets.QVBoxLayout()
		self.verticalLayout.setContentsMargins(0, 0, 0, 0)
		self.verticalLayout.setSpacing(0)
		self.setLayout(self.verticalLayout)
		
		self.verticalLayout.addWidget(self.header)
		self.verticalLayout.addWidget(self.body)
		
		self.header.horizontalScrollBar().valueChanged.connect(self.on_horizontalScroll)
		self.body.horizontalScrollBar().valueChanged.connect(self.on_horizontalScroll)
		self.header.horizontalHeader().sectionResized.connect(self.on_sectionResized)

	def on_horizontalScroll(self, position):
		
		self.header.horizontalScrollBar().setSliderPosition(position)
		self.body.horizontalScrollBar().setSliderPosition(position)
		
	def on_sectionResized(self, index, oldSize, newSize):
		
		self.body.horizontalHeader().resizeSection(index, newSize)

class ExternalHeader(QtWidgets.QTableWidget):
	
	def __init__(self, externalsource, sheet):
		
		QtWidgets.QTableWidget.__init__(self)
		
		self._externalsource = externalsource
		self._sheet = sheet
		
		self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
		self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
		self.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
		self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
		self.horizontalHeader().setStretchLastSection(True)
		self.horizontalHeader().resizeSections(QtWidgets.QHeaderView.ResizeToContents)
		self.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignLeft)
		self.verticalHeader().setVisible(False)
		
		columns_n = self._externalsource.column_count(self._sheet)
		self.setRowCount(1)
		self.setColumnCount(columns_n)
		for idx in range(columns_n):
			name = self._externalsource.column_name(self._sheet, idx)
			# set column label
			item = QtWidgets.QTableWidgetItem()
			item.setData(QtCore.Qt.DisplayRole, name)
			self.setHorizontalHeaderItem(idx, item)
			
			# set target
			item = QtWidgets.QTableWidgetItem()
			item.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsEditable)
			item.setData(QtCore.Qt.DisplayRole, name)
			self.setItem(0, idx, item)
		
		self.adjustSize()
	
	def sheet_name(self):
		
		return self._sheet
		
class ExternalBody(QtWidgets.QTableWidget):
	
	def __init__(self, externalframe, externalsource, sheet = "0"):
		
		QtWidgets.QTableWidget.__init__(self)
		
		self._externalframe = externalframe
		self._externalsource = externalsource
		self._sheet = sheet
		
		self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
		self.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
		self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
		self.horizontalHeader().setStretchLastSection(True)
		self.horizontalHeader().resizeSections(QtWidgets.QHeaderView.ResizeToContents)
		self.horizontalHeader().setVisible(False)
		self.verticalHeader().setVisible(False)
		
		rows_n = self._externalsource.row_count(self._sheet)
		
		if not rows_n:
			return
		
		columns_n = self._externalsource.column_count(self._sheet)
		
		self.setRowCount(rows_n)
		self.setColumnCount(columns_n)
		
		for row_idx in range(rows_n):
			for column_idx in range(columns_n):
				data = self._externalsource.data(self._sheet, row_idx, column_idx)
				
				item = QtWidgets.QTableWidgetItem()
				item.setData(QtCore.Qt.UserRole, data)
				if isinstance(data, DGeometry):
					item.setData(QtCore.Qt.DisplayRole, data.geometry_type)
					item.setData(QtCore.Qt.DecorationRole, self._externalframe.get_icon("geometry.svg"))
				else:
					item.setData(QtCore.Qt.DisplayRole, data)
				item.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable)
				
				self.setItem(row_idx, column_idx, item)
		
		self.adjustSize()





from deposit_gui.view.vusertools_elements.user_elements import user_controls as UserControls

from PySide2 import (QtWidgets)

class EditorControl(object):
	
	def __init__(self, user_control = None):
		
		if user_control is None:
			user_control = getattr(UserControls, self.__class__.__name__)("", "", "", "")
		self.user_control = user_control
	
	def select_text(self):
		
		return ""

class LineEdit(EditorControl, QtWidgets.QLineEdit):
	
	def __init__(self, user_control = None):
		
		EditorControl.__init__(self, user_control)
		QtWidgets.QLineEdit.__init__(self)
		dclass, descriptor = self.user_control.dclass, self.user_control.descriptor
		if (not dclass) or (not descriptor):
			self.setPlaceholderText("[Class].[Descriptor]")
		else:
			self.setText("[%s].[%s]" % (dclass, descriptor))
		
	def select_text(self):
		
		return self.text().strip()

class PlainTextEdit(EditorControl, QtWidgets.QPlainTextEdit):
	
	def __init__(self, user_control = None):
		
		EditorControl.__init__(self, user_control)
		QtWidgets.QPlainTextEdit.__init__(self)
		self.setFixedHeight(100)
		dclass, descriptor = self.user_control.dclass, self.user_control.descriptor
		if (not dclass) or (not descriptor):
			self.setPlaceholderText("[Class].[Descriptor]")
		else:
			self.setPlainText("[%s].[%s]" % (dclass, descriptor))
		
	def select_text(self):
		
		return self.toPlainText().strip()

class ComboBox(EditorControl, QtWidgets.QComboBox):
	
	def __init__(self, user_control = None):
		
		EditorControl.__init__(self, user_control)
		QtWidgets.QComboBox.__init__(self)
		self.setEditable(True)
		dclass, descriptor = self.user_control.dclass, self.user_control.descriptor
		if (not dclass) or (not descriptor):
			self.lineEdit().setPlaceholderText("[Class].[Descriptor]")
		else:
			self.setCurrentText("[%s].[%s]" % (dclass, descriptor))
	
	def select_text(self):
		
		return self.currentText().strip()

class CheckBox(EditorControl, QtWidgets.QFrame):
	
	def __init__(self, user_control = None):
		
		EditorControl.__init__(self, user_control)
		QtWidgets.QFrame.__init__(self)
		
		self.setLayout(QtWidgets.QHBoxLayout())
		
		dclass, descriptor = self.user_control.dclass, self.user_control.descriptor
		self.select = QtWidgets.QLineEdit()
		if (not dclass) or (not descriptor):
			self.select.setPlaceholderText("[Class].[Descriptor]")
		else:
			self.select.setText("[%s].[%s]" % (dclass, descriptor))
		
		self.checkbox = QtWidgets.QCheckBox()
		self.checkbox.setChecked(True)
		self.checkbox.setEnabled(False)
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().addWidget(self.checkbox)
		self.layout().addWidget(self.select)
		self.layout().addStretch()
	
	def select_text(self):
		
		return self.select.text().strip()

class Select(EditorControl, QtWidgets.QLineEdit):
	
	def __init__(self, user_control = None):
		
		EditorControl.__init__(self, user_control)
		QtWidgets.QLineEdit.__init__(self)
		dclass, descriptor = self.user_control.dclass, self.user_control.descriptor
		if (not dclass) or (not descriptor):
			self.setPlaceholderText("[Class].[Descriptor]")
		else:
			self.setText("[%s].[%s]" % (dclass, descriptor))
	
	def select_text(self):
		
		return self.text().strip()




from deposit_gui.view.vnavigator_elements.query_list import QueryList

from PySide2 import (QtWidgets, QtCore, QtGui)

class QueryWidget(QtWidgets.QWidget):
	
	def __init__(self, vnavigator):
		
		QtWidgets.QWidget.__init__(self, vnavigator)
		
		self._vnavigator = vnavigator
		self._querylist = QueryList(self._vnavigator)
		self._footer = QueryFooter(self._vnavigator)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self.layout().addWidget(self._querylist)
		
		self.layout().addWidget(self._footer)
		
		self.update_footer()
	
	def populate(self, queries):
		
		self._querylist.populate(queries)
	
	def get_selected(self):
		
		return self._querylist.get_selected()
	
	def update_footer(self):
		
		selected = self.get_selected()
		
		self._footer.button_edit.setEnabled(len(selected) == 1)
		self._footer.button_remove.setEnabled(len(selected) > 0)

class QueryFooter(QtWidgets.QFrame):
	
	def __init__(self, vnavigator):

		QtWidgets.QFrame.__init__(self)

		self._vnavigator = vnavigator
		
		self.setLayout(QtWidgets.QHBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().setSpacing(6)
		
		self.layout().addStretch()
		
		self.button_add = QtWidgets.QToolButton(self)
		self.button_add.setIcon(self._vnavigator.get_icon("add_query.svg"))
		self.button_add.setIconSize(QtCore.QSize(32, 32))
		self.button_add.setToolTip("Add Query")
		self.button_add.clicked.connect(self._vnavigator.on_query_add)
		self.layout().addWidget(self.button_add)
		
		self.button_edit = QtWidgets.QToolButton(self)
		self.button_edit.setIcon(self._vnavigator.get_icon("edit_query.svg"))
		self.button_edit.setIconSize(QtCore.QSize(32, 32))
		self.button_edit.setToolTip("Edit Query")
		self.button_edit.clicked.connect(self._vnavigator.on_query_edit)
		self.layout().addWidget(self.button_edit)
		
		self.button_remove = QtWidgets.QToolButton(self)
		self.button_remove.setIcon(self._vnavigator.get_icon("remove_query.svg"))
		self.button_remove.setIconSize(QtCore.QSize(32, 32))
		self.button_remove.setToolTip("Remove Query")
		self.button_remove.clicked.connect(self._vnavigator.on_query_remove)
		self.layout().addWidget(self.button_remove)




'''
Generic class to be re-used also by Deposit in the future
'''

from PySide2 import (QtWidgets, QtCore)
import winreg

class DRegistry(QtCore.QObject):
	
	def __init__(self, app_name):
		
		QtCore.QObject.__init__(self)
		
		self.app_name = app_name
		self._vars = {}
		self.key = None
		self._n_writes = 0
		
		self.key = winreg.CreateKey(winreg.HKEY_CURRENT_USER, "SOFTWARE\%s" % (self.app_name))
		for i in range(winreg.QueryInfoKey(self.key)[1]):
			var, value, _ = winreg.EnumValue(self.key, i)
			self._vars[var] = value
		
		self._set_data = {}
	
	def get(self, var):
		
		if var not in self._vars:
			return ""
#		print("registry get", var, self._vars[var])  # DEBUG crashes if active
		return self._vars[var]
	
	def set(self, var, value):
		
		self._vars[var] = value
		self._set_data[var] = value
		self._n_writes += 1
		
		if self._n_writes > 10:
			self.flush()
	
	def vars(self):
		
		return list(self._vars.keys())
	
	def flush(self):
		
		for var in self._set_data:
			winreg.SetValueEx(self.key, var, 0, winreg.REG_SZ, self._set_data[var])
		self._set_data = {}
		self._n_writes = 0



from deposit_gui.dgui.dconnect_tabs.dconnect_tab_recent import DConnectTabRecent
from deposit_gui.dgui.dconnect_tabs.dconnect_tab_pickle import DConnectTabPickle
from deposit_gui.dgui.dconnect_tabs.dconnect_tab_json import DConnectTabJSON
from deposit_gui.dgui.dconnect_tabs.dconnect_tab_db import DConnectTabDB
from deposit_gui.dgui.dconnect_tabs.dconnect_tab_db_rel import DConnectTabDBRel
from deposit_gui.dgui.dconnect_tabs.dconnect_tab_memory import DConnectTabMemory

from PySide2 import (QtWidgets, QtCore, QtGui)
from pathlib import Path
import os

class DConnectFrame(QtWidgets.QFrame):
	
	signal_clear_recent = QtCore.Signal()
	
	def __init__(self, dialog):
		
		QtWidgets.QFrame.__init__(self)
		
		self._dialog = dialog
		
		self._recent_dir = str(Path.home())
		
		self.setMinimumWidth(600)
		self.setMinimumHeight(400)
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self.setStyleSheet('''
			QPushButton {margin: 3px 3px 3px 3px; padding: 5px 10px 5px 10px;}
		''')
		
		self._tab_recent = DConnectTabRecent(self)
		self._tab_pickle = DConnectTabPickle(self)
		self._tab_json = DConnectTabJSON(self)
		self._tab_db = DConnectTabDB(self)
		self._tab_dbrel = DConnectTabDBRel(self)
		
		self._tabs = QtWidgets.QTabWidget()
		self._tabs.addTab(self._tab_recent, "Recent")
		self._tabs.addTab(self._tab_pickle, "Pickle")
		self._tabs.addTab(self._tab_json, "JSON")
		self._tabs.addTab(self._tab_db, "PostgreSQL")
		self._tabs.addTab(self._tab_dbrel, "PostgreSQL Relational")
		if self.creating_enabled():
			self.tab_memory = DConnectTabMemory(self)
			self._tabs.addTab(self.tab_memory, "Memory")
		
		self.layout().addWidget(self._tabs)
		
		self._dialog.set_title(self.title())
		self._dialog.set_frame(self)
		self._dialog.setModal(True)
	
	def set_recent_dir(self, path):
		
		self._recent_dir = path
	
	def get_recent_dir(self):
		
		return self._recent_dir
	
	def set_recent_connections(self, data):
		# data = [[url], [identifier, connstr], ...]
		
		self._tab_recent.set_recent_connections(data)
		self._tab_db.set_recent_connections(data)
		self._tab_dbrel.set_recent_connections(data)
		if self._tab_recent.has_items():
			self._tabs.setCurrentIndex(0)
	
	def on_connect(self, identifier = None, connstr = None, url = None, datasource = None):
		
		self._dialog.set_data(dict(
			identifier = identifier,
			connstr = connstr,
			url = url,
			datasource = datasource,
		))
		
		self._dialog.accept()
	
	def title(self):
		# re-implement
		
		return "Connect"
	
	def creating_enabled(self):
		# re-implement
		
		return False
	
	def connect_caption(self):
		# re-implement
		
		return "Connect"
	
	def logo(self):
		# re-implement
		
		return None




from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.view.view import View

from PySide2 import (QtWidgets, QtCore, QtGui)
from pathlib import Path
import os

class CView(AbstractSubcontroller):
	
	def __init__(self, cmain, cnavigator, cmdiarea) -> None:
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._view = View(cnavigator._vnavigator, cmdiarea._vmdiarea)
		
		self.progress = self._view.progress
		
		self._view._close_callback = self.cmain.on_close

	def show(self):
		
		self._view.show()
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def get_default_folder(self):
		
		folder = self._view.get_recent_dir()
		if folder:
			return folder
		
		if self.cmain.cmodel.has_local_folder():
			return self.cmain.cmodel.get_folder()
		
		return str(Path.home())
	
	def get_save_path(self, caption, filter):
		# returns path, format
		
		path, format = QtWidgets.QFileDialog.getSaveFileName(self._view, dir = self.get_default_folder(), caption = caption, filter = filter)
		
		return path, format
	
	def get_load_path(self, caption, filter):
		
		path, format = QtWidgets.QFileDialog.getOpenFileName(self._view, dir = self.get_default_folder(), caption = caption, filter = filter)
		
		return path, format
	
	def get_logging_path(self):
		
		return self._view.logging.get_log_path()
	
	def get_existing_folder(self, caption):
		
		folder = QtWidgets.QFileDialog.getExistingDirectory(self._view, dir = self.get_default_folder(), caption = caption)
		
		return folder
	
	def get_recent_dir(self):
		
		return self._view.get_recent_dir()
	
	
	def set_title(self, title):
		
		self._view.set_title(title)
	
	def set_recent_dir(self, path):
		
		if os.path.isfile(path):
			path = os.path.dirname(path)
		if not os.path.isdir(path):
			return
		self._view.set_recent_dir(path)
	
	def set_status_message(self, text):
		
		self._view.statusbar.message(text)
	
	def log_message(self, text):
		
		self._view.logging.append(text)
	
	def show_information(self, caption, text):
		
		QtWidgets.QMessageBox.information(self._view, caption, text)
	
	def show_warning(self, caption, text):
		
		QtWidgets.QMessageBox.warning(self._view, caption, text)
	
	def show_question(self, caption, text):
		
		reply = QtWidgets.QMessageBox.question(self._view, caption, text)
		
		return reply == QtWidgets.QMessageBox.Yes
	
	def close(self):
		
		self._view.close()




from PySide2 import (QtWidgets, QtCore, QtGui)

class DConnectTabMemory(QtWidgets.QFrame):
	
	def __init__(self, parent):
		
		QtWidgets.QFrame.__init__(self, parent)
		
		self.parent = parent
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(10, 10, 10, 10)
		
		label = QtWidgets.QLabel("Create a new database in memory.\n\n")
		label.setAlignment(QtCore.Qt.AlignCenter)
		
		self.connect_button = QtWidgets.QPushButton("Create")
		self.connect_button.clicked.connect(self.on_connect)
		
		self.layout().addStretch()
		self.layout().addWidget(label)
		self.layout().addWidget(self.connect_button, alignment = QtCore.Qt.AlignCenter)
		self.layout().addStretch()
	
	@QtCore.Slot()
	def on_connect(self):
		
		self.parent.on_connect(datasource = "Memory")




from deposit_gui.dialogs.dialog_connect import DialogConnect

class DialogImportSource(DialogConnect):
	
	def title(self):
		
		return "Select Data Source to Import"
	
	def creating_enabled(self):
		
		return False
	
	def logo(self):
		
		return None


from PySide2 import (QtWidgets, QtCore, QtGui)

class DNotification(QtWidgets.QFrame):
	
	def __init__(self, parent):
		
		QtWidgets.QFrame.__init__(self)
		
		self._parent = parent
		self._delay = 2000
		self._queue = []
		
		self.label = QtWidgets.QLabel()
		
		self.setWindowFlags(QtCore.Qt.ToolTip)
		self.setFrameShape(QtWidgets.QFrame.StyledPanel)
		self.setStyleSheet('''
		QFrame { background-color: white;}
		QToolButton {
			margin: 0px 0px 0px 0px;
			border: none;
		}
		QLabel { 
			margin: 0px 10px 10px 10px; 
			font-size: 14px;
		}
		''')
		self.label.setOpenExternalLinks(True)
		
		close_button = QtWidgets.QToolButton()
		close_button.setIcon(QtWidgets.QApplication.style().standardIcon(
			QtWidgets.QStyle.SP_TitleBarCloseButton)
		)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(2, 2, 2, 2)
		self.layout().setSpacing(0)
		self.layout().addWidget(close_button)
		self.layout().addWidget(self.label)
		self.layout().setAlignment(close_button, QtCore.Qt.AlignRight)
		
		close_button.clicked.connect(self.on_close)
		
		self.hide_timer = QtCore.QTimer(singleShot = True, timeout = self.on_hide_timer)
		self.hide()
	
	def set_delay(self, delay):
		
		self._delay = delay
	
	def show(self, text, delay = None):
		
		self._queue.append((text, delay))
		if not self.hide_timer.isActive():
			self.show_from_queue()
	
	def show_from_queue(self):
		
		if not self._queue:
			return
		text, delay = self._queue.pop(0)
		if delay is not None:
			self._delay = delay
		self.hide_timer.stop()
		self.label.setText(text)
		self.adjustSize()
		pos = self._parent.mapToGlobal(self._parent.contentsRect().topLeft())
		size = self.rect().size()
		self.move(pos + QtCore.QPoint(5, 5))
		QtWidgets.QFrame.show(self)
		self.hide_timer.start(self._delay)
	
	def hide(self):
		
		self.hide_timer.stop()
		QtWidgets.QFrame.hide(self)
	
	@QtCore.Slot()
	def on_hide_timer(self):
		
		if self.frameRect().contains(self.mapFromGlobal(QtGui.QCursor.pos())):
			self.hide_timer.start(self._delay)
		else:
			self.hide()
			self.show_from_queue()
	
	@QtCore.Slot()
	def on_close(self):
		
		self._queue = []
		self.hide_timer.stop()
		self.hide()




from deposit_gui.dgui.abstract_subview import AbstractSubview
from deposit_gui.view.vnavigator_elements.class_widget import ClassWidget
from deposit_gui.view.vnavigator_elements.query_widget import QueryWidget

from PySide2 import (QtWidgets, QtCore, QtGui)

class VNavigator(AbstractSubview, QtWidgets.QFrame):
	
	signal_widget_activated = QtCore.Signal()
	
	signal_class_selected = QtCore.Signal(list)					# [DClass, ...]
	signal_class_activated = QtCore.Signal(object, bool, object)# DClass, is_descriptor, parent DClass
	signal_class_up = QtCore.Signal(object, object)				# DClass, DClass above
	signal_class_down = QtCore.Signal(object, object)			# DClass, DClass below
	signal_class_add = QtCore.Signal(object, bool, object)		# DClass, is_descriptor, parent DClass
	signal_class_add_descriptor = QtCore.Signal(object, bool, object)	# DClass, is_descriptor, parent DClass
	signal_class_del_descriptor = QtCore.Signal(object, bool, object)	# DClass, is_descriptor, parent DClass
	signal_class_rename = QtCore.Signal(object, bool, object)	# DClass, is_descriptor, parent DClass
	signal_class_remove = QtCore.Signal(list)					# [DClass, ...]
	
	signal_query_activated = QtCore.Signal(str)					# title
	signal_query_add = QtCore.Signal()
	signal_query_edit = QtCore.Signal(str)						# title
	signal_query_remove = QtCore.Signal(list)					# [title, ...]
	
	def __init__(self) -> None:
		
		AbstractSubview.__init__(self)
		QtWidgets.QFrame.__init__(self)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)
		
		toolbox = QtWidgets.QToolBox()
		toolbox.layout().setSpacing(0)
		toolbox.setStyleSheet('''
			QToolBox {
				icon-size: 32px;
			}
			QToolBox::tab {
				font: bold;
			}
		''')
		
		self.layout().addWidget(toolbox)
		
		self._classwidget = ClassWidget(self)
		self._querywidget = QueryWidget(self)
		
		toolbox.addItem(self._classwidget, self.get_icon("class.svg"), "Classes")
		toolbox.addItem(self._querywidget, self.get_icon("query.svg"), "Queries")
	
	def populate_classes(self, classes):
		
		self._classwidget.populate(classes)
	
	def populate_queries(self, queries):
		
		self._querywidget.populate(queries)
	
	
	def deselect_all(self):
		
		self._classwidget.deselect_all()
	
	def _get_one_selected_class(self):
		# returns class, is_descriptor, parent_class
		
		cls = self._classwidget.get_selected()
		if len(cls) == 1:
			is_descriptor = self._classwidget.get_selected_is_descriptor()
			parent = self._classwidget.get_selected_parent()
			return cls[0], is_descriptor, parent
		return None, None, None
	
	
	@QtCore.Slot(list)
	def on_class_selected(self, classes):
		
		self._classwidget.update_footer()
		self.signal_class_selected.emit(classes)
	
	@QtCore.Slot(object, bool, object)
	def on_class_activated(self, cls, is_descriptor, parent_cls):
		
		self.signal_class_activated.emit(cls, is_descriptor, parent_cls)
	
	@QtCore.Slot()
	def on_widget_activated(self):
		
		self.signal_widget_activated.emit()
	
	@QtCore.Slot()
	def on_class_up(self):
		
		cls, is_descriptor, parent_cls = self._get_one_selected_class()
		if cls:
			above, _ = self._classwidget.get_items_around_selected()
			if above:
				self._classwidget.select_one_above()
				self.signal_class_up.emit(cls, above)
	
	@QtCore.Slot()
	def on_class_down(self):
		
		cls, is_descriptor, parent_cls = self._get_one_selected_class()
		if cls:
			_, below = self._classwidget.get_items_around_selected()
			if below:
				self._classwidget.select_one_below()
				self.signal_class_down.emit(cls, below)
	
	@QtCore.Slot()
	def on_class_add(self):
		cls, is_descriptor, parent_cls = self._get_one_selected_class()
		self.signal_class_add.emit(cls, is_descriptor, parent_cls)
	
	@QtCore.Slot()
	def on_class_add_descriptor(self):
		
		cls, is_descriptor, parent_cls = self._get_one_selected_class()
		self.signal_class_add_descriptor.emit(cls, is_descriptor, parent_cls)
	
	@QtCore.Slot()
	def on_class_del_descriptor(self):
		
		cls, is_descriptor, parent_cls = self._get_one_selected_class()
		self.signal_class_del_descriptor.emit(cls, is_descriptor, parent_cls)
	
	@QtCore.Slot()
	def on_class_rename(self):
		
		cls, is_descriptor, parent_cls = self._get_one_selected_class()
		if cls:
			self.signal_class_rename.emit(cls, is_descriptor, parent_cls)
	
	@QtCore.Slot()
	def on_class_remove(self):
		
		classes = self._classwidget.get_selected()
		if classes:
			self.signal_class_remove.emit(classes)
	
	
	@QtCore.Slot(list)
	def on_query_selected(self, queries):
		
		self._querywidget.update_footer()
	
	@QtCore.Slot(str)
	def on_query_activated(self, title):
		
		self.signal_query_activated.emit(title)
	
	@QtCore.Slot()
	def on_query_add(self):
		
		self.signal_query_add.emit()
	
	@QtCore.Slot()
	def on_query_edit(self):
		
		selected = self._querywidget.get_selected()
		if selected:
			self.signal_query_edit.emit(selected[0])
	
	@QtCore.Slot()
	def on_query_remove(self):
		
		selected = self._querywidget.get_selected()
		if selected:
			self.signal_query_remove.emit(selected)




from deposit_gui.dgui.abstract_subview import AbstractSubview

from PySide2 import (QtWidgets, QtCore, QtGui)

class DialogImportExternal(AbstractSubview, QtWidgets.QFrame):
	
	def __init__(self, dialog, n_rows, classes, relations):
		
		QtWidgets.QFrame.__init__(self)
		AbstractSubview.__init__(self)
		
		self._classes = classes
		self._checkboxes = []
		self._relcontrols = {}
		
		dialog.set_title("Import External Data")
		dialog.setModal(True)
		dialog.set_button_box(True, True)
		
		self.setMinimumWidth(256)
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		frame = QtWidgets.QFrame()
		frame.setLayout(QtWidgets.QHBoxLayout())
		frame.layout().setContentsMargins(0, 0, 0, 0)
		
		grid = QtWidgets.QFrame()
		self.layout().addWidget(frame)
		frame.layout().addWidget(grid)
		frame.layout().addStretch()
		
		grid.setLayout(QtWidgets.QGridLayout())
		grid.layout().addWidget(QtWidgets.QLabel("<b>Rows:</b>"), 0, 0)
		grid.layout().addWidget(QtWidgets.QLabel("%d" % (n_rows)), 0, 1)
		grid.layout().addWidget(QtWidgets.QLabel(), 1, 0)
		grid.layout().addWidget(QtWidgets.QLabel("<b>Class</b>"), 2, 0)
		grid.layout().addWidget(QtWidgets.QLabel("<b>Unique</b>"), 2, 1)
		idx = 0
		for idx, name in enumerate(self._classes):
			self._checkboxes.append(QtWidgets.QCheckBox())
			self._checkboxes[-1]._name = name
			grid.layout().addWidget(QtWidgets.QLabel(name), idx + 3, 0)
			grid.layout().addWidget(self._checkboxes[-1], idx + 3, 1)
		grid.layout().addWidget(QtWidgets.QLabel(), idx + 4, 0)
		grid.layout().addWidget(QtWidgets.QLabel("<b>Relations:</b>"), idx + 5, 0)
		
		self.form = QtWidgets.QFrame()
		self.form.setLayout(QtWidgets.QGridLayout())
		margins = self.form.layout().contentsMargins()
		margins.setTop(0)
		self.form.layout().setContentsMargins(margins)
		self.layout().addWidget(self.form)
		
		row = 0
		for src, label, tgt in relations:
			ctrls = []
			ctrls.append(QtWidgets.QComboBox())
			ctrls[-1].addItems(self._classes)
			ctrls[-1].setCurrentIndex(self._classes.index(src))
			ctrls[-1].setEditable(True)
			ctrls.append(QtWidgets.QLineEdit(label))
			ctrls.append(QtWidgets.QComboBox())
			ctrls[-1].addItems(self._classes)
			ctrls[-1].setCurrentIndex(self._classes.index(tgt))
			ctrls[-1].setEditable(True)
			ctrls.append(RemoveButton(row, self.get_icon("trash.svg")))
			ctrls[-1].signal_clicked.connect(self.on_remove)
			
			self._relcontrols[row] = ctrls
			
			self.form.layout().addWidget(ctrls[0], row, 0)
			self.form.layout().addWidget(ctrls[1], row, 1)
			self.form.layout().addWidget(ctrls[2], row, 2)
			self.form.layout().addWidget(ctrls[3], row, 3)
			
			row += 1
		
		button_add = QtWidgets.QToolButton()
		button_add.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
		button_add.setIcon(self.get_icon("link.svg"))
		button_add.setIconSize(QtCore.QSize(24,24))
		button_add.setText("Add")
		button_add.clicked.connect(self.on_add)
		
		frame = QtWidgets.QFrame()
		frame.setLayout(QtWidgets.QHBoxLayout())
		frame.layout().addStretch()
		frame.layout().addWidget(button_add)
		frame.layout().addStretch()
		self.layout().addWidget(frame)
		
	def add_row(self):
		
		row = self.form.layout().rowCount()
		
		ctrls = []
		ctrls.append(QtWidgets.QComboBox())
		ctrls[-1].addItems([""] + self._classes)
		ctrls[-1].setEditable(True)
		ctrls.append(QtWidgets.QLineEdit(""))
		ctrls.append(QtWidgets.QComboBox())
		ctrls[-1].addItems([""] + self._classes)
		ctrls[-1].setEditable(True)
		ctrls.append(RemoveButton(row, self.get_icon("trash.svg")))
		ctrls[-1].signal_clicked.connect(self.on_remove)
		
		self._relcontrols[row] = ctrls
		
		self.form.layout().addWidget(ctrls[0], row, 0)
		self.form.layout().addWidget(ctrls[1], row, 1)
		self.form.layout().addWidget(ctrls[2], row, 2)
		self.form.layout().addWidget(ctrls[3], row, 3)
	
	@QtCore.Slot(int)
	def on_remove(self, row):
		
		for col in range(4):
			item = self.form.layout().itemAtPosition(row, col)
			item.widget().deleteLater()
		del self._relcontrols[row]
	
	@QtCore.Slot()
	def on_add(self):
		
		self.add_row()
	
	def get_unique(self):
		
		unique = set()
		for checkbox in self._checkboxes:
			if checkbox.isChecked():
				unique.add(checkbox._name)
		return unique
	
	def get_relations(self):
		
		relations = set()
		for row in self._relcontrols:
			combo_src, edit_label, combo_tgt, _ = self._relcontrols[row]
			src = combo_src.currentText().strip()
			tgt = combo_tgt.currentText().strip()
			label = edit_label.text().strip()
			if label and (src in self._classes) and (tgt in self._classes):
				relations.add((src, label, tgt))
		return relations



class RemoveButton(QtWidgets.QToolButton):
	
	signal_clicked = QtCore.Signal(int)
	
	def __init__(self, row, icon):
		
		QtWidgets.QToolButton.__init__(self)
		
		self.setIcon(icon)
		self.setToolTip("Remove Relation")
		
		self._row = row
		self.clicked.connect(self.on_clicked)
	
	@QtCore.Slot()
	def on_clicked(self):
		
		self.signal_clicked.emit(self._row)
		


from deposit_gui.dgui.dconnect_frame import DConnectFrame
from deposit_gui.dgui.dclickable_logo import DClickableLogo

import deposit_gui

from PySide2 import (QtWidgets, QtCore, QtGui)
import os

class DialogConnect(DConnectFrame):
	
	def __init__(self, dialog):
		
		DConnectFrame.__init__(self, dialog)
	
	def title(self):
		
		return "Select Data Source"
	
	def creating_enabled(self):
		
		return True
	
	def logo(self):
		
		path = os.path.join(os.path.dirname(deposit_gui.__file__), "res/dep_installer.svg")
		
		logo_frame = QtWidgets.QFrame()
		logo_frame.setLayout(QtWidgets.QVBoxLayout())
		logo_frame.layout().setContentsMargins(0, 0, 0, 0)
		logo_frame.layout().addStretch()
		logo_frame.layout().addWidget(DClickableLogo(path, "https://github.com/demjanp/deposit", alignment = QtCore.Qt.AlignCenter))
		logo_frame.layout().addStretch()
		
		return logo_frame



from deposit_gui.dgui.dgraphics_view import DGraphicsView
from deposit_gui.dgui import pygraphviz

from PySide2 import (QtWidgets, QtCore, QtGui, QtPrintSupport)
from collections import defaultdict
import networkx as nx
import weakref
import math
import re

def to_agraph(N):
	
	directed = N.is_directed()
	strict = nx.number_of_selfloops(N) == 0 and not N.is_multigraph()
	A = pygraphviz.AGraph(name=N.name, strict=strict, directed=directed)

	# default graph attributes
	A.graph_attr.update(N.graph.get("graph", {}))
	A.node_attr.update(N.graph.get("node", {}))
	A.edge_attr.update(N.graph.get("edge", {}))

	A.graph_attr.update(
		(k, v) for k, v in N.graph.items() if k not in ("graph", "node", "edge")
	)

	# add nodes
	for n, nodedata in N.nodes(data=True):
		A.add_node(n)
		# Add node data
		a = A.get_node(n)
		a.attr.update({k: str(v) for k, v in nodedata.items()})

	# loop over edges
	if N.is_multigraph():
		for u, v, key, edgedata in N.edges(data=True, keys=True):
			str_edgedata = {k: str(v) for k, v in edgedata.items() if k != "key"}
			A.add_edge(u, v, key=str(key))
			# Add edge data
			a = A.get_edge(u, v)
			a.attr.update(str_edgedata)

	else:
		for u, v, edgedata in N.edges(data=True):
			str_edgedata = {k: str(v) for k, v in edgedata.items()}
			A.add_edge(u, v)
			# Add edge data
			a = A.get_edge(u, v)
			a.attr.update(str_edgedata)

	return A

def pygraphviz_layout(G, prog="dot", root=None, args=""):
	
	if root is not None:
		args += f"-Groot={root}"
	A = to_agraph(G)
	A.layout(prog=prog, args=args)
	node_pos = {}
	for n in G:
		node = pygraphviz.Node(A, n)
		try:
			xs = node.attr["pos"].split(",")
			node_pos[n] = tuple(float(x) for x in xs)
		except:
			print("no position for node", n)
			node_pos[n] = (0.0, 0.0)
	return node_pos


NODE_TYPE = QtWidgets.QGraphicsItem.UserType + 1
EDGE_TYPE = QtWidgets.QGraphicsItem.UserType + 2

class AbstractNode(QtWidgets.QGraphicsItem):
	
	type = NODE_TYPE
	
	def __init__(self, node_id, label = "", font_family = "Calibri", font_size = 16):
		
		self.node_id = node_id
		self.label = str(label)
		self.edges = []  # [Edge, ...]
		self.font = QtGui.QFont(font_family, font_size)
		self.label_w = 0
		self.label_h = 0
		
		QtWidgets.QGraphicsItem.__init__(self)
		
		if self.label != "":
			rect = QtGui.QFontMetrics(self.font).boundingRect(self.label)
			self.label_w = rect.width()
		
		self.setAcceptHoverEvents(True)
		self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable)
		self.setFlag(QtWidgets.QGraphicsItem.ItemSendsGeometryChanges)
		self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable, True)
		self.setFlag(QtWidgets.QGraphicsItem.ItemIsFocusable, True)
		self.setCacheMode(self.DeviceCoordinateCache)
		self.setZValue(-1)
	
	def add_edge(self, edge):
		
		self.edges.append(weakref.ref(edge))
		edge.adjust()
	
	def del_edge(self, edge):
		
		edge = weakref.ref(edge)
		if edge in self.edges:
			self.edges.remove(edge)
	
	def has_parent(self):
		
		for edge in self.edges:
			edge = edge()
			if edge is None:
				continue
			if edge.source() != self:
				return True
		return False
	
	def get_parents(self):
		
		parents = set()
		for edge in self.edges:
			edge = edge()
			if edge is None:
				continue
			node = edge.source()
			if node != self:
				parents.add(node)
		
		return parents
	
	def has_child(self):
		
		for edge in self.edges:
			edge = edge()
			if edge is None:
				continue
			if edge.target() != self:
				return True
		return False
	
	def get_children(self):
		
		children = set()
		for edge in self.edges:
			edge = edge()
			if edge is None:
				continue
			node = edge.target()
			if node != self:
				children.add(node)
		
		return children
	
	def center(self):
		
		return self.boundingRect().center()
	
	def shape(self):
		
		return QtGui.QPainterPath()
	
	def on_hover(self, state):
		# state: True = Enter, False = Leave
		# re-implement
		
		pass
	
	def on_position_change(self):
		# re-implement
		
		pass
	
	def on_mouse_press(self):
		# re-implement
		
		pass
	
	def on_mouse_release(self):
		# re-implement
		
		pass
	
	def hoverEnterEvent(self, event):
		
		self.on_hover(True)
		
		QtWidgets.QGraphicsItem.hoverEnterEvent(self, event)
	
	def hoverLeaveEvent(self, event):
		
		self.on_hover(False)
		
		QtWidgets.QGraphicsItem.hoverLeaveEvent(self, event)
	
	def itemChange(self, change, value):
		
		if change == QtWidgets.QGraphicsItem.ItemPositionChange:
			for edge in self.edges:
				edge = edge()
				if edge is not None:
					edge.adjust()
			
			self.on_position_change()
		
		return QtWidgets.QGraphicsItem.itemChange(self, change, value)
	
	def mousePressEvent(self, event):
		
		self.update()
		self.on_mouse_press()
		QtWidgets.QGraphicsItem.mousePressEvent(self, event)
	
	def mouseReleaseEvent(self, event):
		
		self.update()
		self.on_mouse_release()
		QtWidgets.QGraphicsItem.mouseReleaseEvent(self, event)

class Node(AbstractNode):
	
	def __init__(self, node_id, label = "", radius = 10, font_family = "Calibri", font_size = 16):
		
		AbstractNode.__init__(self, node_id, label, font_family, font_size)
		
		self.radius = radius
	
	def boundingRect(self):
		
		adjust = 2		
		return QtCore.QRectF(-self.radius - adjust, -self.radius - adjust, (2*self.radius) + 3 + adjust, (2*self.radius) + 3 + adjust)
	
	def center(self):
		
		return QtCore.QPointF(0, 0)
	
	def shape(self):
		
		path = QtGui.QPainterPath()
		path.addEllipse(-self.radius, -self.radius, 2*self.radius, 2*self.radius)
		return path
	
	def paint(self, painter, option, widget):
		
		pen_width = 0
		if option.state & QtWidgets.QStyle.State_Sunken:
			pen_width = 2
		if option.state & QtWidgets.QStyle.State_Selected:
			painter.setBrush(QtGui.QBrush(QtCore.Qt.lightGray))
		else:
			painter.setBrush(QtGui.QBrush(QtCore.Qt.white))
		painter.setPen(QtGui.QPen(QtCore.Qt.black, pen_width))
		painter.drawEllipse(-self.radius, -self.radius, 2*self.radius, 2*self.radius)
		if self.label != "":
			painter.setFont(self.font)
			painter.drawText(25, 8, self.label)
	
	def on_hover(self, state):
		# state: True = Enter, False = Leave
		# re-implement
		
		pass
	
	def on_position_change(self):
		# re-implement
		
		pass
	
	def on_mouse_press(self):
		# re-implement
		
		pass
	
	def on_mouse_release(self):
		# re-implement
		
		pass
	
class NodeWithAttributes(AbstractNode):
	
	def __init__(self, node_id, label = "", descriptors = [], font_family = "Calibri", font_size = 14, max_width = 256):
		# descriptors = [(name, value), ...]
		
		AbstractNode.__init__(self, node_id, label, font_family, font_size)
		
		self.descriptors = []
		self.descriptor_w = 0
		self.descriptor_h = 0
		self.selection_polygon = None
		self.selection_shape = None
		self.text_padding = 3
		self.max_width = max_width
		
		self.populate_descriptors(descriptors)
		self.adjust()
	
	def populate_descriptors(self, descriptors):
		
		for name, value in descriptors:
			if value:
				self.descriptors.append("%s: %s" % (name, value))
			else:
				self.descriptors.append(name)
	
	def adjust(self):
		
		adjust_w = 2
		adjust_h = 0
		
		if self.label != "":
			rect = QtGui.QFontMetrics(self.font).boundingRect(self.label)
			self.label_w = rect.width() + adjust_w
			self.label_h = rect.height() + adjust_h
		
		self.descriptor_w = 0
		self.descriptor_h = 0
		for text in self.descriptors:
			rect = QtGui.QFontMetrics(self.font).boundingRect(text)
			self.descriptor_w = max(self.descriptor_w, rect.width() + adjust_w)
			self.descriptor_h = max(self.descriptor_h, rect.height() + adjust_h)
		
		self.descriptor_w = min(self.descriptor_w, self.max_width)
		
		w = max(self.label_w, self.descriptor_w) + 2*self.text_padding
		h = self.label_h + 2*self.text_padding + len(self.descriptors)*(self.descriptor_h + self.text_padding) + self.text_padding
		
		self.selection_polygon = QtGui.QPolygonF(QtCore.QRectF(0, 0, w, h))
		self.selection_shape = QtGui.QPainterPath()
		self.selection_shape.addPolygon(self.selection_polygon)
	
	def boundingRect(self):
		
		return self.selection_polygon.boundingRect()
	
	def center(self):
		
		return self.boundingRect().center()
	
	def shape(self):
		
		return self.selection_shape
	
	def paint(self, painter, option, widget):
		
		adjust_y = -2
		
		pen_width = 0
		if option.state & QtWidgets.QStyle.State_Sunken:
			pen_width = 2
		
		y = self.label_h + 2*self.text_padding + adjust_y
		rect = self.selection_polygon.boundingRect()
		
		painter.setPen(QtCore.Qt.NoPen)
		painter.setBrush(QtGui.QBrush(QtCore.Qt.white))
		painter.drawPath(self.selection_shape)
		
		painter.setPen(QtCore.Qt.NoPen)
		if option.state & QtWidgets.QStyle.State_Selected:
			painter.setBrush(QtGui.QBrush(QtCore.Qt.darkGray))
		else:
			painter.setBrush(QtGui.QBrush(QtCore.Qt.lightGray))
		painter.drawRect(0, 0, rect.width(), y)
		
		painter.setPen(QtGui.QPen(QtCore.Qt.black, pen_width))
		painter.setBrush(QtCore.Qt.NoBrush)
		painter.drawLine(0, y, rect.width(), y)
		painter.drawPath(self.selection_shape)
		
		painter.setFont(self.font)
		
		if self.label != "":
			painter.drawText(self.text_padding, self.label_h + adjust_y, self.label)
		
		for n, text in enumerate(self.descriptors):
			painter.drawText(self.text_padding, y + (n + 1)*(self.descriptor_h + self.text_padding) + adjust_y, text)
	
	def on_hover(self, state):
		# state: True = Enter, False = Leave
		# re-implement
		
		pass
	
	def on_position_change(self):
		# re-implement
		
		pass
	
	def on_mouse_press(self):
		# re-implement
		
		pass
	
	def on_mouse_release(self):
		# re-implement
		
		pass

class NodeWithSimpleAttributes(NodeWithAttributes):
	
	def populate_descriptors(self, descriptors):
		
		for _, value in descriptors:
			self.descriptors.append(str(value))
	
	def adjust(self):
		
		adjust_w = 2
		adjust_h = 0
		
		self.descriptor_w = 0
		self.descriptor_h = 0
		for text in self.descriptors:
			rect = QtGui.QFontMetrics(self.font).boundingRect(text)
			self.descriptor_w = max(self.descriptor_w, rect.width() + adjust_w)
			self.descriptor_h = max(self.descriptor_h, rect.height() + adjust_h)
		
		self.descriptor_w = min(self.descriptor_w, self.max_width)
		
		w = self.descriptor_w + 2*self.text_padding
		h = len(self.descriptors)*(self.descriptor_h + self.text_padding) + self.text_padding
		
		self.selection_polygon = QtGui.QPolygonF(QtCore.QRectF(0, 0, w, h))
		self.selection_shape = QtGui.QPainterPath()
		self.selection_shape.addPolygon(self.selection_polygon)
	
	def paint(self, painter, option, widget):
		
		adjust_y = -2
		
		pen_width = 0
		if option.state & QtWidgets.QStyle.State_Sunken:
			pen_width = 2
		
		painter.setPen(QtCore.Qt.NoPen)
		if option.state & QtWidgets.QStyle.State_Selected:
			painter.setBrush(QtGui.QBrush(QtCore.Qt.lightGray))
		else:
			painter.setBrush(QtGui.QBrush(QtCore.Qt.white))
		
		painter.setPen(QtGui.QPen(QtCore.Qt.black, pen_width))
		
		painter.drawPath(self.selection_shape)
		
		y = adjust_y
		
		painter.setFont(self.font)
		
		for n, text in enumerate(self.descriptors):
			painter.drawText(self.text_padding, y + (n + 1)*(self.descriptor_h + self.text_padding) + adjust_y, text)
	
	def on_hover(self, state):
		# state: True = Enter, False = Leave
		# re-implement
		
		pass
	
	def on_position_change(self):
		# re-implement
		
		pass
	
	def on_mouse_press(self):
		# re-implement
		
		pass
	
	def on_mouse_release(self):
		# re-implement
		
		pass

class Edge(QtWidgets.QGraphicsItem):
	
	type = EDGE_TYPE
	
	def __init__(self, source, target, label = "", color = None, offset = 0, font_family = "Calibri", font_size = 16, arrow_size = 20):
		# offset = display edge as curve with highest point offset by this distance from a line
		
		QtWidgets.QGraphicsItem.__init__(self)
		
		self.label = label
		self.color = QtCore.Qt.gray if color is None else color
		self.source = weakref.ref(source)
		self.target = weakref.ref(target)
		self.offset = offset
		self.is_loop = (source.node_id == target.node_id)
		self.arrow_size = arrow_size
		self.font = QtGui.QFont(font_family, font_size)
		self.label_w = 0
		self.label_h = 0
		self.line = None
		self.edge_path = QtGui.QPainterPath()
		self.selection_shape = QtGui.QPainterPath()
		self._line_width = None
		
		if self.label != "":
			rect = QtGui.QFontMetrics(self.font).boundingRect(self.label)
			self.label_w = rect.width()
			self.label_h = rect.height()
		
		if isinstance(self.color, str):
			self.color = getattr(QtCore.Qt, self.color)
		
		self.setFlag(QtWidgets.QGraphicsItem.ItemSendsGeometryChanges)
		self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable, True)
		self.setFlag(QtWidgets.QGraphicsItem.ItemIsFocusable, True)
		self.setCacheMode(self.DeviceCoordinateCache)
		self.setZValue(-2)
		
		self.source().add_edge(self)
		self.target().add_edge(self)
		self.adjust()
	
	def set_line_width(self, value):
		
		self._line_width = value
	
	def get_angle(self):
		
		if self.is_loop:
			return 45
		length = self.line.length()
		if length:
			angle = math.acos(self.line.dx() / length)
		else:
			angle = 0
		if self.line.dy() >= 0:
			angle = 2*math.pi - angle
		return angle
	
	def adjust(self):
		
		self.edge_path = QtGui.QPainterPath()
		self.selection_shape = QtGui.QPainterPath()
		
		if not self.source() or not self.target():
			return
		
		self.line = QtCore.QLineF(self.mapFromItem(self.source(), self.source().center()), self.mapFromItem(self.target(), self.target().center()))
		
		self.prepareGeometryChange()
		
		selection_offset = 7
		
		if self.is_loop:
			d = self.arrow_size * 3 if (self.arrow_size > 0) else 3
			x, y = self.line.p1().x() - d, self.line.p1().y() - d
			self.edge_path.arcMoveTo(x, y, d, d, 270)
			self.edge_path.arcTo(x, y, d, d, 270, 360)
			self.selection_shape = QtGui.QPainterPath(self.edge_path)
			
		else:
			angle = self.get_angle()
			self.edge_path.moveTo(self.line.p1())
			offset = ((self.offset + 1) // 2)
			step = max(10, self.arrow_size*2.5, self.label_h*2)
			offset *= step
			if self.offset % 2 == 1:
				offset *= -1
			controlpoint = self.line.center() + QtCore.QPointF(math.sin(angle - math.pi) * offset, math.cos(angle - math.pi) * offset)
			self.edge_path.quadTo(controlpoint, self.line.p2())
			
			rad_angle = self.line.angle() * math.pi / 180
			dx = selection_offset * math.sin(rad_angle)
			dy = selection_offset * math.cos(rad_angle)
			path = QtGui.QPainterPath(self.edge_path)
			path.translate(dx, dy)
			self.selection_shape.connectPath(path)
			path = QtGui.QPainterPath(self.edge_path.toReversed())
			path.translate(-dx, -dy)
			self.selection_shape.connectPath(path)
	
	def boundingRect(self):
		
		rect = self.selection_shape.boundingRect()
		x0, y0, w, h = rect.x(), rect.y(), rect.width(), rect.height()
		x1, y1 = x0 + w, y0 + h
		if self.is_loop:
			x0 -= self.arrow_size
		if self.label != "":
			text_pos = self.edge_path.pointAtPercent(0.5)
			x0 = min(x0, text_pos.x() - self.label_w / 2 - 4)
			y0 = min(y0, text_pos.y() - self.label_h / 2 - 4)
			x1 = max(x1, text_pos.x() + self.label_w / 2)
			y1 = max(y1, text_pos.y() + self.label_h / 2)
		return QtCore.QRectF(x0, y0, x1 - x0, y1 - y0)
	
	def shape(self):
		
		return self.selection_shape
	
	def paint(self, painter, option, widget):
		
		if not self.source() or not self.target():
			return
		
		pen_width = 1
		if self._line_width is not None:
			pen_width = self._line_width
		if option.state & QtWidgets.QStyle.State_Selected:
			pen_width *= 2
		
		pen = QtGui.QPen(self.color, pen_width, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin)
		pen.setCosmetic(True)
		
		painter.setPen(pen)
		
		angle = self.get_angle()
		
		painter.strokePath(self.edge_path, pen)
		
		if self.arrow_size > 0:
			arrow_pos = self.edge_path.pointAtPercent(0.6)
			arrow_p1 = arrow_pos + QtCore.QPointF(math.sin(angle - math.pi / 3) * self.arrow_size, math.cos(angle - math.pi / 3) * self.arrow_size)
			arrow_p2 = arrow_pos + QtCore.QPointF(math.sin(angle - math.pi + math.pi / 3) * self.arrow_size, math.cos(angle - math.pi + math.pi / 3) * self.arrow_size)
			painter.setBrush(self.color)
			painter.drawPolygon(QtGui.QPolygonF([arrow_pos, arrow_p1, arrow_p2]))
		
		if self.label != "":
			painter.setFont(self.font)
			painter.setPen(QtGui.QPen(QtCore.Qt.black, pen_width))
			text_pos = self.edge_path.pointAtPercent(0.5)
			x = text_pos.x()
			y = text_pos.y()
			painter.drawText(x - self.label_w / 2, y + self.label_h / 4, self.label)

class DGraphView(DGraphicsView):
	
	signal_node_activated = QtCore.Signal(object)
	signal_selected = QtCore.Signal()
	
	def __init__(self):
		
		self._nodes = {}  # {node_id: Node, ...}
		self._edges = []  # [Edge, ...]
		self._mouse_prev = None
		self._edge_font_family = 16
		self._edge_font_size = 16
		self._search_edit = None
		
		DGraphicsView.__init__(self)
		
		self.scene().setItemIndexMethod(QtWidgets.QGraphicsScene.NoIndex)
		self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorViewCenter)
		self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
		
		self.setMinimumSize(200, 200)
		
		self.scene().selectionChanged.connect(self.on_selected)
	
	def show_search_box(self, text = "Search", icon = None):
		
		if self._search_edit is not None:
			return
		self._search_edit = QtWidgets.QLineEdit()
		if text:
			self._search_edit.setPlaceholderText(text)
		self._search_edit.setFixedWidth(100)
		self._search_edit.textChanged.connect(self.on_search)
		search_box = QtWidgets.QFrame()
		search_box.setLayout(QtWidgets.QHBoxLayout())
		if icon is not None:
			label = QtWidgets.QLabel()
			label.setPixmap(icon.pixmap(QtCore.QSize(24, 24)))
			search_box.layout().addWidget(label)
		search_box.layout().addWidget(self._search_edit)
		self.layout().addWidget(search_box)
		self.layout().setAlignment(search_box, QtCore.Qt.AlignLeft)
	
	def clear(self):
		
		DGraphicsView.clear(self)
		
		self._nodes.clear()
		self._edges.clear()
	
	def calc_positions(self, nodes, edges, gap = 10, x_stretch = 1, y_stretch = 1, move_nodes = None):
		# nodes = [AbstractNode, node_id, ...]
		# edges = [[source_id, target_id], ...]
		# move_nodes = [node_id, ...]; if set, only calculate positions for those nodes
		#
		# returns {node_id: (x, y), ...}
		
		pos_collect = {}
		done = set([])
		y_max = 0
		if move_nodes is not None:
			done = set(move_nodes)
		G = nx.DiGraph()
		for row in edges:
			source_id, target_id = row[:2]
			if move_nodes is not None:
				if (source_id not in move_nodes) or (target_id not in move_nodes):
					continue
			G.add_edge(source_id, target_id)
			if move_nodes is None:
				done.add(source_id)
				done.add(target_id)
		for node_id in nodes:
			if isinstance(node_id, AbstractNode):
				node_id = node_id.node_id
			if (move_nodes is not None) and (node_id not in move_nodes):
				continue
			G.add_node(node_id)
			if move_nodes is None:
				done.add(node_id)
		for node_id in done:
			if node_id not in self._nodes:
				continue
			rect = self._nodes[node_id].boundingRect()
			mul = 0.013  # TODO find a way to calculate
			w, h = rect.width(), rect.height()
			G.nodes[node_id]["width"] = (w + 2*gap) * mul
			G.nodes[node_id]["height"] = (h + 2*gap) * mul
		
		g_positions = pygraphviz_layout(G, prog = "dot")
		
		for node_id in g_positions:
			x, y = g_positions[node_id]
			g_positions[node_id] = (x * x_stretch, y * y_stretch)
		xmin, ymax = math.inf, -math.inf
		for node_id in g_positions:
			x, y = g_positions[node_id]
			xmin = min(xmin, x)
			ymax = max(ymax, y)
		for node_id in g_positions:
			x, y = g_positions[node_id]
			g_positions[node_id] = (x - xmin, ymax - y)
		
		for node_id in g_positions:
			pos_collect[node_id] = tuple(g_positions[node_id])
			y_max = max(y_max, pos_collect[node_id][1])
		
		n = len([node_id for node_id in self._nodes if node_id not in done])
		if n > 3:
			n_side = max(1, int(n**0.5))
		else:
			n_side = n
		
		x = 0
		y = y_max
		h_max = 0
		n = 0
		for node_id in self._nodes:
			if node_id in done:
				continue
			pos_collect[node_id] = (x, y)
			rect = self._nodes[node_id].boundingRect()
			w, h = rect.width() + gap, rect.height() + gap
			h_max = max(h_max, h)
			x += w
			n += 1
			if n > n_side:
				x = 0
				y += h_max
				h_max = 0
				n = 0
		
		return pos_collect
	
	def populate(self, nodes = [], edges = [], positions = {}, gap = 10, x_stretch = 1, y_stretch = 1, edge_font_family = "Calibri", edge_font_size = 16, edge_arrow_size = 20, move_nodes = None, progress = None):
		# nodes = [AbstractNode, (node_id, label), ...]
		# edges = [[source_id, target_id], [source_id, target_id, label], [source_id, target_id, label, color], ...]
		# positions = {node_id: (x, y), ...}
		# move_nodes = [node_id, ...]; if set, only calculate positions for those nodes
		
		self.clear()
		
		self.hide()
		
		for node in nodes:
			if isinstance(node, tuple):
				node_id, label = node
				node = Node(node_id, label)
			self._nodes[node.node_id] = node
		
		node_ids = move_nodes
		if not node_ids:
			node_ids = list(self._nodes.keys())
		if False in [(node_id in positions) for node_id in node_ids]:
			positions_new = self.calc_positions(self._nodes, edges, gap, x_stretch, y_stretch, move_nodes)
			positions_new.update(positions)
			positions.clear()
			positions.update(positions_new)
		
		cnt = 0
		cmax = len(self._nodes) + len(edges)
		for node_id in self._nodes:
			cnt += 1
			if progress is not None:
				progress.update_state(text = "Drawing Tree", value = cnt, maximum = cmax)
				if progress.cancel_pressed():
					self.clear()
					self.show()
					return
			
			self._nodes[node_id].setPos(*positions[node_id])
			self.scene().addItem(self._nodes[node_id])
		
		edges_done = defaultdict(int)
		for row in edges:
			cnt += 1
			if progress is not None:
				progress.update_state(text = "Drawing Tree", value = cnt, maximum = cmax)
				if progress.cancel_pressed():
					self.clear()
					self.show()
					return
			attrs = [None, None, "", "black"]
			attrs[:len(row)] = row
			source_id, target_id, label, color = attrs
			if (source_id not in self._nodes) or (target_id not in self._nodes):
				continue
			key = tuple(sorted([source_id, target_id]))
			self._edges.append(Edge(
				self._nodes[source_id], self._nodes[target_id], label, color, 
				offset = edges_done[key], 
				font_family = edge_font_family, 
				font_size = edge_font_size, 
				arrow_size = edge_arrow_size
			))
			edges_done[key] += 1
			self.scene().addItem(self._edges[-1])
		
		self.show()
		self.reset_zoom()
	
	def add_node(self, node, position = None):
		# node = AbstractNode
		# position = (x, y)
		
		self._nodes[node.node_id] = node
		if position is not None:
			self._nodes[node.node_id].setPos(*position)
		self.scene().addItem(self._nodes[node.node_id])
	
	def add_edge(self, source_id, target_id, label = "", color = None, offset = 0, font_family = "Calibri", font_size = 16, arrow_size = 20):
		
		if (source_id not in self._nodes) or (target_id not in self._nodes):
			return
		
		self._edges.append(Edge(self._nodes[source_id], self._nodes[target_id], label, color, offset, font_family, font_size, arrow_size))
		self.scene().addItem(self._edges[-1])
	
	def del_node(self, node_id):
		
		if node_id not in self._nodes:
			return
		self.scene().removeItem(self._nodes[node_id])
		del self._nodes[node_id]
		to_del = []
		for edge in self._edges:
			if (edge.source().node_id == node_id) or (edge.target().node_id == node_id):
				self.scene().removeItem(edge)
				to_del.append(edge)
		for edge in to_del:
			self._edges.remove(edge)
	
	def del_edge(self, source_id, target_id):
		
		to_del = []
		for edge in self._edges:
			if (edge.source().node_id == source_id) and \
				(edge.target().node_id == target_id):
					to_del.append(edge)
		for edge in to_del:
			edge.source().del_edge(edge)
			edge.target().del_edge(edge)
			self._edges.remove(edge)
			self.scene().removeItem(edge)
	
	def del_edges(self, node_ids):
		# node_ids = [node_id, ...]
		
		node_ids = set(node_ids)
		to_del = []
		for edge in self._edges:
			if node_ids.intersection([edge.source().node_id, edge.target().node_id]):
				to_del.append(edge)
		for edge in to_del:
			edge.source().del_edge(edge)
			edge.target().del_edge(edge)
			self._edges.remove(edge)
			self.scene().removeItem(edge)
	
	def get_node(self, node_id):
		
		if node_id in self._nodes:
			return self._nodes[node_id]
		return None
	
	def has_nodes(self):
		
		return len(self._nodes) > 0
	
	def get_nodes(self):
		
		for node_id in self._nodes:
			yield self._nodes[node_id]
	
	def get_edges(self):
		
		return self._edges
	
	def get_selected(self):
		
		nodes = []
		edges = []
		for item in self.scene().selectedItems():
			if item.type == NODE_TYPE:
				nodes.append(item.node_id)
			elif item.type == EDGE_TYPE:
				edges.append((item.source().node_id, item.target().node_id, item.label))
		return nodes, edges
	
	def get_positions(self):
		
		positions = {}
		for node_id in self._nodes:
			pos = self._nodes[node_id].pos()
			x, y = pos.x(), pos.y()
			positions[node_id] = (x, y)
		return positions
	
	def get_scale(self):
		
		return self.transform().m11()
	
	def select_node(self, node_id):
		
		if node_id in self._nodes:
			self._nodes[node_id].setSelected(True)
	
	def select_all(self):
		
		for node_id in self._nodes:
			self._nodes[node_id].setSelected(True)
	
	def deselect_all(self):
		
		self.scene().clearSelection()
	
	def save_pdf(self, path, dpi = 72, page_size = QtGui.QPageSize.A4, stroke_width = None):
		
		self.scene().clearSelection()
		
		scale = self.get_scale()
		
		if stroke_width is not None:
			for item in self.scene().items():
				if isinstance(item, Edge):
					item.set_line_width(stroke_width)
		
		rect = self.scene().itemsBoundingRect()
		m = min(rect.width(), rect.height())*0.05
		rect = rect.marginsAdded(QtCore.QMarginsF(m, m, m, m))
		w, h = rect.width(), rect.height()
		
		printer = QtPrintSupport.QPrinter()
		printer.setWinPageSize(page_size)
		printer.setFullPage(True)
		is_landscape = False
		if w > h:
			is_landscape = True
			printer.setOrientation(QtPrintSupport.QPrinter.Landscape)
		else:
			printer.setOrientation(QtPrintSupport.QPrinter.Portrait)
		printer.setOutputFormat(QtPrintSupport.QPrinter.PdfFormat)
		printer.setOutputFileName(path)
		size = printer.pageLayout().pageSize().sizePoints()
		if is_landscape:
			pw, ph = size.height(), size.width()
		else:
			pw, ph = size.width(), size.height()
		scale = min(pw / w, ph / h)
		printer.setResolution(int(round(dpi)))
		pw *= scale
		ph *= scale
		if is_landscape:
			printer.setPageSize(QtGui.QPageSize(QtCore.QSize(ph, pw), units = QtGui.QPageSize.Point))
		else:
			printer.setPageSize(QtGui.QPageSize(QtCore.QSize(pw, ph), units = QtGui.QPageSize.Point))
		
		painter = QtGui.QPainter(printer)
		self.scene().render(painter, source = rect)
		painter.end()
		
		if stroke_width is not None:
			for item in self.scene().items():
				if isinstance(item, Edge):
					item.set_line_width(None)
	
	@QtCore.Slot()
	def on_selected(self):
		
		self.signal_selected.emit()
	
	@QtCore.Slot(str)
	def on_search(self, text):
		
		node = None
		for node_id in self._nodes:
			if re.search(text.strip(), str(self._nodes[node_id].label), re.IGNORECASE):
				node = self._nodes[node_id]
				break
		if node is None:
			return
		pos = node.pos()
		scale = self.get_scale()
		self.scale_view((1 / scale) * 0.6)
		self.centerOn(pos)
		self.deselect_all()
		self.select_node(node.node_id)
	
	def mousePressEvent(self, event):
		
		if event.button() == QtCore.Qt.LeftButton:
			self.setDragMode(QtWidgets.QGraphicsView.RubberBandDrag)
			self.setCursor(QtCore.Qt.ArrowCursor)
		elif event.button() == QtCore.Qt.RightButton:
			self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
			self.setCursor(QtCore.Qt.OpenHandCursor)
			self._mouse_prev = (event.x(), event.y())
			return
		
		QtWidgets.QGraphicsView.mousePressEvent(self, event)
	
	def mouseMoveEvent(self, event):
		
		if self._mouse_prev is not None:
			prev_point = self.mapToScene(*self._mouse_prev)
			new_point = self.mapToScene(event.pos())
			translation = new_point - prev_point
			self.translate(translation.x(), translation.y())
			self._mouse_prev = (event.x(), event.y())
		
		QtWidgets.QGraphicsView.mouseMoveEvent(self, event)
	
	def mouseReleaseEvent(self, event):
		
		self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
		self.setCursor(QtCore.Qt.ArrowCursor)
		self._mouse_prev = None
		
		QtWidgets.QGraphicsView.mouseReleaseEvent(self, event)
	
	def mouseDoubleClickEvent(self, event):
		
		item = self.itemAt(event.pos())
		if item is None:
			return
		if item.type == NODE_TYPE:
			self.signal_node_activated.emit(item.node_id)




from deposit_gui.dgui.abstract_subview import AbstractSubview

from PySide2 import (QtWidgets, QtCore, QtGui)

class VQueryToolbar(AbstractSubview):
	
	signal_entered = QtCore.Signal(str)
	
	def __init__(self, vmain) -> None:
		
		AbstractSubview.__init__(self, vmain)
		
		self.vmain.addToolBarBreak()
		self._toolbar = self.vmain.addToolBar("Query")
		
		self._query_box = QtWidgets.QComboBox()
		self._query_box.setEditable(True)
		self._query_box.setSizePolicy(
			QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed
		)
		self._query_box.setMinimumHeight(26)
		self._query_box.completer().setCaseSensitivity(QtCore.Qt.CaseSensitive)
		
		query_button = QtWidgets.QToolButton()
		query_button.setText("Enter")
		query_button.setIcon(self.vmain.get_icon("next_small.svg"))
		query_button.setStyleSheet('''
		QToolButton {
			margin: 0px 0px 0px 0px;
			width: 22px;
			height: 22px;
		}
		''')
		query_button.clicked.connect(self.on_enter)
		
		self._toolbar.addWidget(QtWidgets.QLabel("Query: "))
		self._toolbar.addWidget(self._query_box)
		self._toolbar.addWidget(query_button)
		
		self._query_box.lineEdit().returnPressed.connect(self.on_enter)
	
	def set_query_text(self, text):
		
		self._query_box.setCurrentText(text)
	
	def get_query_text(self):
		
		return self._query_box.currentText().strip()
	
	@QtCore.Slot()
	def on_enter(self):
		
		text = self._query_box.currentText().strip()
		if text:
			self.signal_entered.emit(text)




from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.dgui.dvactions import DVActions

from PySide2 import (QtCore)

class DCActions(AbstractSubcontroller):
	
	def __init__(self, cmain, cview) -> None:
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._vactions = DVActions(cview._view)
		
		self._vactions.toolbar.signal_triggered.connect(self.on_toolbar_triggered)
		self._vactions.toolbar.signal_combo_changed.connect(self.on_toolbar_combo_changed)
		self._vactions.menubar.signal_triggered.connect(self.on_menubar_triggered)
		
		self._vactions.toolbar.populate(self.set_up_tool_bar())
		self._vactions.menubar.populate(self.set_up_menu_bar())
	
	def update(self):
		
		for bar in [self._vactions.toolbar, self._vactions.menubar]:
			for name in bar.get_names():
				if hasattr(self, "update_%s" % (name)):
					bar.update_tool(name, getattr(self, "update_%s" % (name))())
	
	@QtCore.Slot(str)
	def on_toolbar_triggered(self, name):
		
		if hasattr(self, "on_%s" % (name)):
			getattr(self, "on_%s" % (name))(self._vactions.toolbar.get_state(name))
	
	@QtCore.Slot(str, str)
	def on_toolbar_combo_changed(self, name, text):
		
		if hasattr(self, "on_%s" % (name)):
			getattr(self, "on_%s" % (name))(text)
	
	@QtCore.Slot(str)
	def on_menubar_triggered(self, name):
		
		if hasattr(self, "on_%s" % (name)):
			getattr(self, "on_%s" % (name))(self._vactions.menubar.get_state(name))
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def get_combo_value(self, name):
		
		return self._vactions.toolbar.get_combo_value(name)
		
	def set_combo_value(self, name, text):
		
		self._vactions.toolbar.set_combo_value(name, text)
	
	
	# ---- Actions
	# ------------------------------------------------------------------------
	def set_up_tool_bar(self):
		# re-implement to fill tool bar
		
		return {}
		'''
		return {
			"Toolbar": [
				("Name", "Caption"),
				("Name", "Caption"),
			],
		}
		'''
	
	def set_up_menu_bar(self):
		# re-implement to fill menu bar
		
		return {}
		'''
		return {
			"Menu": [
				("Name", "Caption"),
				("Name", "Caption"),
			],
		}
		'''
	
	
	# implement update_[name] and on_[name] for each action
	'''
	def update_[name](self):
		
		return dict(
			caption = "Caption",
			icon = "icon.svg",
			shortcut = "Ctrl+S",
			help = "Tool tip",
			combo: list,
			checkable = True,
			checked = True,
			enabled = True,
		)
	
	def on_[name](self, state / text):
		
		pass
	'''



from deposit_gui.view.vusertools_elements.user_elements import user_groups
from deposit_gui.view.vusertools_elements.editor.editor_frame import (EditorFrame)

from PySide2 import (QtWidgets, QtCore)

class EditorGroup(QtWidgets.QGroupBox):
	
	def __init__(self, element, form_editor, user_group = None):
		
		self.label_edit = None
		self.element = element
		self.group = None
		self.hovered = False
		self.selected = False
		self.bold = False
		self.form_editor = form_editor
		
		QtWidgets.QGroupBox.__init__(self, self.element)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(10, 10, 10, 10)
		
		if user_group is None:
			user_group = getattr(user_groups, element)("", "Label")
		self.group = user_group
		
		self.label_edit = QtWidgets.QLineEdit()
		if not self.group.label:
			self.label_edit.setPlaceholderText("Label")
		else:
			self.label_edit.setText(self.group.label)
		form = QtWidgets.QFrame()
		form.setLayout(QtWidgets.QFormLayout())
		form.layout().setContentsMargins(0, 0, 0, 0)
		form.layout().addRow("Label:", self.label_edit)
		self.layout().addWidget(form)
		
		self.controls_frame = QtWidgets.QFrame()
		self.controls_frame.setLayout(QtWidgets.QVBoxLayout())
		self.controls_frame.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().addWidget(self.controls_frame)
		
		for member in self.group.members:
			self.add_frame(member.__class__.__name__, member)
		
		self.bold = "font-weight: bold;" in self.group.stylesheet
		
		self.update_stylesheet()
		
		self.setMouseTracking(True)
	
	def add_frame(self, element, user_control = None, before = None):
		
		if before is None:
			self.controls_frame.layout().addWidget(EditorFrame(element, self, user_control))
		else:
			idx = self.controls_frame.layout().indexOf(before)
			self.controls_frame.layout().insertWidget(idx, EditorFrame(element, self, user_control))
	
	def remove_control(self, element):
		
		self.controls_frame.layout().removeWidget(element)
	
	def user_element(self):
		
		label = self.label_edit.text()
		if not label:
			return None
		self.group.label = label
		self.group.stylesheet = "QGroupBox {font-weight: bold;}" if self.bold else ""
		self.group.members = []
		for element in self.controls_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
			if isinstance(element, EditorFrame):
				self.group.members.append(element.user_element())
		return self.group
		
	def get_selected(self):
		
		for element in self.controls_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
			if isinstance(element, EditorFrame):
				if element.selected:
					return element
		return None
		
	def deselect_all(self):
		
		for element in self.controls_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
			if isinstance(element, EditorFrame):
				element.setSelected(False)
	
	def update_stylesheet(self):
		
		stylesheet = ""
		if self.hovered:
			stylesheet += "%s {background: lightgrey;}" % (self.__class__.__name__)
		if self.selected:
			stylesheet += " %s {border: 2px solid grey;}" % (self.__class__.__name__)
		if self.bold:
			stylesheet += " %s {font-weight: bold;}" % (self.__class__.__name__)
		if stylesheet:
			self.setStyleSheet(stylesheet)
	
	def setSelected(self, state):
		
		self.selected = state
		self.update_stylesheet()
		self.form_editor.on_selection_changed()
	
	def setBold(self, state):
		
		self.bold = state
		self.update_stylesheet()
	
	def setParent(self, parent):
		
		if parent is None:
			for element in self.findChildren(QtWidgets.QWidget):
				element.setParent(None)
		QtWidgets.QGroupBox.setParent(self, parent)
	
	def enterEvent(self, event):
		
		self.hovered = True
		self.update_stylesheet()
	
	def leaveEvent(self, event):
		
		self.hovered = False
		self.update_stylesheet()
		
	def mousePressEvent(self, event):
		
		state = not self.selected
		self.form_editor.deselect_all()
		self.setSelected(state)
		self.form_editor.on_selection_changed()




from deposit.utils.fnc_files import (
	is_local_url, url_to_path, get_temp_path, get_named_path, get_image_format,
	get_free_subfolder, open_url, get_updated_local_url
)
from deposit_gui.utils.fnc_svg import (svg_to_raster)

from PySide2 import (QtCore, QtGui, QtSvg)
from PIL import Image
import hashlib
import shutil
import os

def make_thumbnail(src_path, tgt_path, size, src_format):
	# generate a thumbnail
	
	stat = os.stat(src_path)
	ta, tm = stat.st_atime, stat.st_mtime
	
	if src_format == "svg":
		svg_to_raster(
			src_path, tgt_path, max_side = size, stroke_width = size * 0.004
		)
	else:
		try:
			icon = Image.open(src_path).convert("RGB")
		except:
			return
		icon.thumbnail((size, size), Image.ANTIALIAS)
		icon.save(tgt_path, quality = 100)
		icon.close()
	
	os.utime(tgt_path, (ta, tm))


def get_thumbnail(resource, local_folder, cache, size = 256):
	# resource = DResource
	# find or make thumbnail from image specified by resource
	# return path
	
	# update local url if local_folder has moved
	url, filename, is_stored, is_image = resource.value
	if is_stored:
		url_new = get_updated_local_url(url, local_folder)
		if url_new is None:
			return None
		if url_new != url:
			resource.value = (url_new, filename, is_stored, is_image)
			url = url_new
	
	# thumbnail filename
	tmb_filename = "%s_%s.jpg" % (hashlib.md5(filename.encode("utf-8")).hexdigest(), size)
	
	if is_local_url(url):
		img_path = url_to_path(url)
		tm_orig = os.stat(img_path).st_mtime
	else:
		img_path = None
		tm_orig = -1
	
	if tmb_filename in cache:
		# check if thumbnails modified time matches the source file
		if (tm_orig == -1) or (tm_orig == os.stat(cache[tmb_filename]).st_mtime):
			return cache[tmb_filename]
	
	local_folder = get_named_path("_thumbnails", local_folder)
	tgt_path = os.path.join(get_free_subfolder(local_folder), tmb_filename)
	
	src_format = get_image_format(url)
	if src_format is None:
		return None
	
	src_path = os.path.join(get_temp_path("thumbnails"), os.path.splitext(tmb_filename)[0] + src_format)
	
	# make temporary copy of image
	f_src = open_url(url)
	if f_src is None:
		return None
	f_tgt = open(src_path, "wb")
	shutil.copyfileobj(f_src, f_tgt)
	f_src.close()
	f_tgt.close()
	
	if tm_orig > -1:
		# set copys modified time to the originals
		ta = os.stat(src_path).st_atime
		os.utime(src_path, (ta, tm_orig))
	
	# generate new thumbnail
	make_thumbnail(src_path, tgt_path, size, src_format)
	
	cache[tmb_filename] = tgt_path
	
	return tgt_path


def load_thumbnails(local_folder, cache):
	
	local_folder = os.path.join(local_folder, "_thumbnails")
	if os.path.exists(local_folder):
		for dirname in os.listdir(local_folder):
			dirname = os.path.join(local_folder, dirname)
			if os.path.isdir(dirname):
				for filename in os.listdir(dirname):
					cache[filename] = os.path.abspath(os.path.join(dirname, filename))







from deposit.utils.fnc_files import (as_url)
from deposit import Store

from PySide2 import (QtWidgets, QtCore, QtGui)
import os

class AbstractTabFile(QtWidgets.QFrame):
	
	EXTENSION = ""  # json, pickle etc.
	DATASOURCE = None
	
	def __init__(self, parent):
		
		QtWidgets.QFrame.__init__(self, parent)
		
		self.parent = parent
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		homedir = self.parent.get_recent_dir()
		
		self.fs_model = QtWidgets.QFileSystemModel()
		self.fs_model.setRootPath(homedir)
		self.fs_model.setNameFilters(["*.%s" % self.EXTENSION])
		self.fs_model.setNameFilterDisables(False)
		self.tree = FileTree()
		self.tree.setModel(self.fs_model)
		for i in range(1, self.fs_model.columnCount()):
			self.tree.hideColumn(i)
		self.tree.setCurrentIndex(self.fs_model.index(homedir))
		self.tree.setExpanded(self.fs_model.index(homedir), True)
		self.tree.setAnimated(False)
		self.tree.setIndentation(20)
		self.tree.setSortingEnabled(True)
		self.tree.selected.connect(self.on_selected)
		self.tree.activated.connect(self.on_connect)
		
		self.path_edit = QtWidgets.QLineEdit()
		self.path_edit.textChanged.connect(self.on_path_changed)
		
		self.connect_button = QtWidgets.QPushButton(self.parent.connect_caption())
		self.connect_button.clicked.connect(self.on_connect)
		
		self.layout().addWidget(self.tree)
		self.layout().addWidget(self.path_edit)
		self.layout().addWidget(self.connect_button)
		
		self.update()

	def get_path(self):
		
		path = self.path_edit.text().strip()
		if (not path) or os.path.isdir(path):
			return ""
		path, filename = os.path.split(path)
		filename, ext = os.path.splitext(filename)
		if ext.lower() != ".%s" % self.EXTENSION:
			ext = ".%s" % self.EXTENSION
		path = os.path.join(path, "%s%s" % (filename, ext))
		return path
	
	def update(self):
		
		path = self.get_path()
		is_valid = False
		if os.path.isfile(path):
			self.connect_button.setText(self.parent.connect_caption())
			is_valid = True
		elif self.parent.creating_enabled():
			self.connect_button.setText("Create")
			is_valid = True
		self.connect_button.setEnabled(is_valid and (path != ""))
	
	@QtCore.Slot()
	def on_selected(self):
		
		index = self.tree.currentIndex()
		path = self.fs_model.filePath(index)
		self.path_edit.setText(path)
	
	@QtCore.Slot()
	def on_path_changed(self):
		
		self.update()
	
	@QtCore.Slot()
	def on_connect(self):
		
		path = self.get_path()
		if not path:
			return
		
		if (not os.path.isfile(path)) and (not self.parent.creating_enabled()):
			QtWidgets.QMessageBox.critical(self, "Error", "Could not create database.")
			return None
		
		if not os.path.isfile(path):
			store = Store(keep_temp = True)
			store.save(path = path)
		
		url = as_url(path)
		self.parent.on_connect(url = url)

class FileTree(QtWidgets.QTreeView):
	
	selected = QtCore.Signal()
	
	def selectionChanged(self, selected, deselected):
		
		self.selected.emit()
		
		QtWidgets.QTreeView.selectionChanged(self, selected, deselected)




from deposit_gui.view.vmdiarea_frames.abstract_mdiarea_frame import AbstractMDIAreaFrame
from PySide2 import (QtWidgets, QtCore, QtGui)
from natsort import (natsorted)
from urllib.parse import (urlencode, parse_qsl)

class RelationFrame(QtWidgets.QFrame):
	
	signal_object_link = QtCore.Signal(int)		# obj_id
	signal_class_link = QtCore.Signal(str)		# class_name
	signal_relation_link = QtCore.Signal(int, str, str)	# obj_id, rel_label, class_name
	signal_relation_unlink = QtCore.Signal(int, str, str)	# obj_id, rel_label, class_name
	
	def __init__(self):
		
		QtWidgets.QFrame.__init__(self)
		
		self._obj = None
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		self.setStyleSheet('''
			QFrame {
				background-color: white;
			}
		''')
		
		form_header = QtWidgets.QWidget()
		form_header.setLayout(QtWidgets.QFormLayout())
		form_header.layout().setContentsMargins(0, 0, 0, 0)
		self.label_obj_id = QtWidgets.QLabel("None")
		self.label_obj_id.setTextInteractionFlags(QtCore.Qt.TextBrowserInteraction)
		self.label_obj_id.linkActivated.connect(self.on_object_link)
		self.label_classes = QtWidgets.QLabel("None")
		self.label_classes.setTextInteractionFlags(QtCore.Qt.TextBrowserInteraction)
		self.label_classes.linkActivated.connect(self.on_class_link)
		form_header.layout().addRow("<b>Object ID:</b>", self.label_obj_id)
		form_header.layout().addRow("<b>Classes:</b>", self.label_classes)
		
		self.list_related = QtWidgets.QWidget()
		self.list_related.setLayout(QtWidgets.QFormLayout())
		self.list_related.layout().setContentsMargins(0, 0, 0, 0)
		
		self.layout().addWidget(form_header)
		self.layout().addWidget(QtWidgets.QLabel("<b>Related:</b>"))
		self.layout().addWidget(self.list_related)
		self.layout().addStretch()
		
	def populate(self, obj = None):
		
		if obj is None:
			obj = self._obj
		
		self._obj = obj
		
		for i in reversed(range(self.list_related.layout().rowCount())):
			self.list_related.layout().removeRow(i)
		
		if obj is None:
			self.label_obj_id.setText("None")
			self.label_classes.setText("None")
			return
		
		class_names = [cls.name for cls in obj.get_classes(ordered = True)]
		
		self.label_obj_id.setText(
			"<a style=\"text-decoration:none;\" href=\"%s\">%d</a>" % \
				(urlencode({"obj_id": obj.id}), obj.id)
		)
		self.label_classes.setText(", ".join([
			"<a style=\"text-decoration:none;\" href=\"%s\">%s</a>" % \
				(urlencode({"name": name}), name) for name in class_names
		]))
		relations = set()
		for obj_tgt, label in obj.get_relations():
			for cls in obj_tgt.get_classes():
				relations.add((label, cls.name))
		relations = natsorted(list(relations))
		for label, name in relations:
			label_link = QtWidgets.QLabel(
				"<a style=\"text-decoration:none;\" href=\"%s\">%s.%s</a>" % \
					(urlencode({"obj_id": obj.id, "label": label, "name": name}), label, name)
			)
			label_link.setTextInteractionFlags(QtCore.Qt.TextBrowserInteraction)
			label_link.linkActivated.connect(self.on_relation_link)
			
			label_unlink = QtWidgets.QLabel(
				"<a style=\"text-decoration:none;\" href=\"%s\">unlink</a>" % \
					(urlencode({"obj_id": obj.id, "label": label, "name": name}))
			)
			label_unlink.setTextInteractionFlags(QtCore.Qt.TextBrowserInteraction)
			label_unlink.linkActivated.connect(self.on_relation_unlink)
			
			self.list_related.layout().addRow(label_link, label_unlink)
	
	@QtCore.Slot(str)
	def on_object_link(self, text):
		
		obj_id = int(parse_qsl(text)[0][1])
		self.signal_object_link.emit(obj_id)
	
	@QtCore.Slot(str)
	def on_class_link(self, text):
		
		name = parse_qsl(text)[0][1]
		self.signal_class_link.emit(name)
	
	@QtCore.Slot(str)
	def on_relation_link(self, text):
		
		data = dict(parse_qsl(text))
		obj_id = int(data["obj_id"])
		label = data["label"]
		name = data["name"]
		self.signal_relation_link.emit(obj_id, label, name)
	
	@QtCore.Slot(str)
	def on_relation_unlink(self, text):
		
		data = dict(parse_qsl(text))
		obj_id = int(data["obj_id"])
		label = data["label"]
		name = data["name"]
		self.signal_relation_unlink.emit(obj_id, label, name)
		




from PySide2 import (QtWidgets, QtCore, QtGui)

class UserTool(QtWidgets.QAction):
	
	def __init__(self, label, vusertools):
		
		self.label = label
		self._vusertools = vusertools
		
		QtWidgets.QAction.__init__(self, self._vusertools.vmain)
		
		self.setText(self.label)
		self.triggered.connect(self.on_triggered)
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		pass
	
	def to_markup(self):
		
		return '''
<Title>%s</>
<Type %s/>
		''' % (self.label, self.__class__.__name__)
	
	def to_dict(self):
		
		return dict(
			typ = self.__class__.__name__,
			label = self.label,
		)




from deposit_gui.view.vmdiarea_frames.query_frame_elements.abstract_query_tab import AbstractQueryTab
from deposit_gui.view.vmdiarea_frames.query_frame_elements.abstract_drag_model import AbstractDragModel
from deposit_gui.view.vmdiarea_frames.query_frame_elements.query_item import QueryItem

from PySide2 import (QtWidgets, QtCore, QtGui)

class ImageDelegate(QtWidgets.QStyledItemDelegate):
	
	def __init__(self, query_tab_images):
		
		QtWidgets.QStyledItemDelegate.__init__(self, query_tab_images)
		
		self._query_tab_images = query_tab_images
	
	def paint(self, painter, option, index):
		
		self._query_tab_images._list_model.on_paint(index.data(QtCore.Qt.UserRole))
		
		QtWidgets.QStyledItemDelegate.paint(self, painter, option, index)

class ProxyModel(QtCore.QSortFilterProxyModel):
	
	def __init__(self):
		
		super(ProxyModel, self).__init__()
		
		self._item_order = None
	
	def lessThan(self, source_left, source_right):
		
		def get_order(source):
			
			if self._item_order is None:
				return 0
			item = source.data(QtCore.Qt.UserRole)
			if item is None:
				return 0
			if item._key in self._item_order:
				return self._item_order[item._key]
			return 0
		
		return get_order(source_left) < get_order(source_right)
	
	def filterAcceptsRow(self, row, parent):
		
		if self._item_order is None:
			return True
		item = self.sourceModel().index(row, 0, parent).data(QtCore.Qt.UserRole)
		if item is None:
			return False
		if item._key in self._item_order:
			return True
		return False
	
	def sort(self, item_order):
		
		self._item_order = item_order
		
		QtCore.QSortFilterProxyModel.sort(self, 0, QtCore.Qt.DescendingOrder)
		QtCore.QSortFilterProxyModel.sort(self, 0, QtCore.Qt.AscendingOrder)
	
	def filter(self, item_order):
		
		self._item_order = item_order
		
		self.invalidateFilter()

class ListModel(AbstractDragModel, QtCore.QAbstractListModel):
	
	icon_loaded = QtCore.Signal(QtCore.QModelIndex)
	
	def __init__(self, queryframe, images, item_order, cmodel, icon_size = 256):
		
		QtCore.QAbstractListModel.__init__(self)
		
		self._queryframe = queryframe
		self._query = queryframe._query
		self._cmodel = cmodel
		self._icon_size = icon_size
		self._proxy_model = ProxyModel()
		
		pixmap = QtGui.QPixmap(self._icon_size, self._icon_size)
		pixmap.fill()
		self.empty_icon = QtGui.QIcon(pixmap)
		
		self._proxy_model.setSourceModel(self)
		
		self.populate_data(images, item_order)
	
	def populate_data(self, images, item_order):
		
		self._icons = {} # {(row, col): QIcon or None, ...}; for each image
		self._images = images # {(row, col): DResource, ...}
		self._keys = sorted(list(images.keys()))
		self._item_order = item_order  # {(row, col): order, ...}
		
		self._proxy_model.sort(self._item_order)
	
	def rowCount(self, parent):
		
		return len(self._keys)
	
	def flags(self, index):
		
		item = self.data(index, QtCore.Qt.UserRole)
		
		return AbstractDragModel.flags(self, item)
	
	def data(self, index, role):
		
		key = self._keys[index.row()]
		
		if role == QtCore.Qt.DisplayRole:
			return self._images[key].filename
		
		if role == QtCore.Qt.DecorationRole:
			icon = self._icons.get(key, None)
			if icon is None:
				return self.empty_icon
			return icon
		
		if role == QtCore.Qt.UserRole:
			class_name, descriptor_name = self._query.columns[key[1]]
			value = self._images[key]
			obj_id = list(value.object_ids)
			obj_id = obj_id[0] if obj_id else None
			item = QueryItem(index, class_name, descriptor_name, obj_id, value)
			item._key = key
			
			return item
		
		return None
	
	def drag_supported(self, item):
		# item: QueryItem
		
		return True
	
	def drop_supported(self, item):
		# item: QueryItem
		
		return True
	
	def on_drop_url(self, item, urls):
		# item: QueryItem
		
		if urls:
			self._queryframe.on_drop_url(item, urls[0])
		
	def on_drop_text(self, item, text):
		# item: QueryItem
		
		return
		
	def on_drop_items(self, tgt_item, items):
		# tgt_item: QueryItem
		# items: [QueryItem, ...]
		
		if items:
			self._queryframe.on_edited(tgt_item, items[0].value)
	
	def on_paint(self, item):
		
		if item is None:
			return
		
		key = item._key
		if self._icons.get(key, None) is None:
			path = self._cmodel.get_thumbnail(item.value, self._icon_size)
			if path is not None:
				self._icons[key] = QtGui.QIcon(path)
				self.icon_loaded.emit(self._proxy_model.mapFromSource(item.index))
		
	def on_close(self):
		
		pass

class ListView(AbstractQueryTab, QtWidgets.QListView):
	
	def __init__(self, queryframe, header, images, item_order, cmodel, icon_size = 256):
		
		QtWidgets.QListView.__init__(self)
		
		self._queryframe = queryframe
		self._query = queryframe._query
		self._header = header
		self._icon_size = icon_size
		self._thumbnail_size = icon_size
		
		self.setAcceptDrops(True)
		self.setDragEnabled(True)
		self.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
		self.setDefaultDropAction(QtCore.Qt.CopyAction)
		self.setDropIndicatorShown(True)
		
		self.setItemDelegate(ImageDelegate(self))
		self._list_model = ListModel(queryframe, images, item_order, cmodel, self._icon_size)
		
		self.setViewMode(QtWidgets.QListView.IconMode)
		self.setUniformItemSizes(True)
		self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
		self.setResizeMode(QtWidgets.QListView.Adjust)
		self.setWrapping(True)
		self.setFlow(QtWidgets.QListView.LeftToRight)
		
		self.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
		
		self.setModel(self._list_model._proxy_model)
		
		self._list_model.icon_loaded.connect(self.update)
		
		self.activated.connect(self.on_activated)
	
	def update_query(self, images, item_order):
		
		self._list_model.beginResetModel()
		self._list_model.populate_data(images, item_order)
		self._list_model.endResetModel()
	
	def get_row_count(self):
		
		return self._list_model._proxy_model.rowCount()
	
	def set_thumbnail_size(self, value):
		
		self._thumbnail_size = value
		self.setIconSize(QtCore.QSize(value, value))
	
	def sort(self, item_order):
		
		self._list_model._proxy_model.sort(item_order)
	
	def apply_filter(self, item_order):
		
		self._list_model._proxy_model.filter(item_order)
	
	@QtCore.Slot(QtCore.QModelIndex)
	def on_activated(self, index):
		
		item = index.data(QtCore.Qt.UserRole)
		self._queryframe.on_query_activated(item)
	
	def on_selected(self):
		
		items = [index.data(QtCore.Qt.UserRole) for index in self.selectionModel().selectedIndexes()]
		items = [item for item in items if item is not None]
		self._queryframe.on_query_selected(items)
		self._header.set_to_object_enabled(len(items) == 1)
		if items:
			self._header.set_object(items[0].obj_id)
		else:
			self._header.set_object(None)
	
	def on_close(self):
		
		self._list_model.on_close()
	
	def selectionChanged(self, *args, **kwargs):
		
		QtWidgets.QListView.selectionChanged(self, *args, **kwargs)
		
		self.on_selected()

class ListHeader(QtWidgets.QFrame):
	
	def __init__(self, queryframe):
		
		QtWidgets.QFrame.__init__(self)
		
		self._queryframe = queryframe
		self._obj_id = None
		
		self.setFrameShape(QtWidgets.QFrame.StyledPanel)
		self.setFrameShadow(QtWidgets.QFrame.Raised)
		self.setLayout(QtWidgets.QHBoxLayout())
		self.layout().setContentsMargins(5, 0, 0, 0)
		self.layout().setSpacing(0)
		
		self.to_object_button = QtWidgets.QToolButton()
		self.to_object_button.setIcon(self._queryframe.get_icon("to_object.svg"))
		self.to_object_button.setIconSize(QtCore.QSize(24, 24))
		self.to_object_button.setAutoRaise(True)
		self.to_object_button.setToolTip("Select Object")
		self.to_object_button.setContentsMargins(0, 0, 10, 0)
		self.to_object_button.setEnabled(False)
		self.to_object_button.clicked.connect(lambda state: self._queryframe.on_to_object(self._obj_id))
		self.layout().addWidget(self.to_object_button)
		
		zoom_label = QtWidgets.QLabel("Zoom:")
		zoom_label.setContentsMargins(5, 0, 5, 0)
		self.layout().addWidget(zoom_label)
		self.zoom_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
		self.zoom_slider.setMinimum(64)
		self.zoom_slider.setMaximum(256)
		self.zoom_slider.setMaximumWidth(400)
		self.zoom_slider.valueChanged.connect(self._queryframe.on_zoom)
		self.layout().addWidget(self.zoom_slider)
		
		self.layout().addStretch()
	
	def set_object(self, obj_id):
		
		self._obj_id = obj_id
	
	def set_to_object_enabled(self, state):
		
		self.to_object_button.setEnabled(state)
	
	def set_zoom_slider(self, value):

		self.blockSignals(True)
		self.zoom_slider.setValue(value)
		self.blockSignals(False)


class QueryTabImagesLazy(AbstractQueryTab, QtWidgets.QWidget):
	
	def __init__(self, queryframe):
		
		QtWidgets.QWidget.__init__(self)
		
		self._queryframe = queryframe
		self._query = queryframe._query
		self._has_image = None
		
	def has_image(self):
		
		if self._has_image is None:
			for row in self._query:
				for _, value in row:
					if (value.__class__.__name__ == "DResource") and value.is_image:
						self._has_image = True
						return True
			self._has_image = False
		return self._has_image
	
	def get_row_count(self):
		
		return int(self.has_image())
	
	def set_thumbnail_size(self, value):
		
		pass
	
	def sort(self, item_order):
		
		pass
	
	def clearSelection(self):
		
		pass

class QueryTabImages(AbstractQueryTab, QtWidgets.QFrame):
	
	def __init__(self, queryframe, images, item_order, cmodel, icon_size = 256):
		
		QtWidgets.QFrame.__init__(self)
		
		self._list_header = ListHeader(queryframe)
		self._list_view = ListView(queryframe, self._list_header, images, item_order, cmodel, icon_size)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().setSpacing(0)
		self.layout().addWidget(self._list_header)
		self.layout().addWidget(self._list_view)
	
	def set_thumbnail_size(self, value):
		
		self._list_header.set_zoom_slider(value)
		self._list_view.set_thumbnail_size(value)
	
	def update_query(self, images, item_order):
		
		self._list_view.update_query(images, item_order)
	
	def get_row_count(self):
		
		return self._list_view.get_row_count()
	
	def apply_filter(self, item_order):
		
		self._list_view.apply_filter(item_order)
	
	def sort(self, item_order):
		
		self._list_view.sort(item_order)
	
	def clearSelection(self):
		
		self._list_view.clearSelection()
	
	def on_selected(self):
		
		self._list_view.on_selected()
	
	def on_close(self):
		
		self._list_view.on_close()

